"""

Corrplot example
==================

"""
# some useful pylab imports for this notebook


# Create some random data
import string
letters = string.ascii_uppercase[0:15]
import pandas as pd
import numpy as np
df = pd.DataFrame(dict(( (k, np.random.random(10)+ord(k)-65) for k in letters)))
df = df.corr()

# if the input is not a square matrix or indices do not match 
# column names, correlation is computed on the fly
from biokit.viz import corrplot
c = corrplot.Corrplot(df)


c.plot(colorbar=False, method='square', shrink=.9 ,rotation=45)





# -*- python -*-
#
#  This file is part of biokit software
#
#  Copyright (c) 2015 - EBI-EMBL
#  Copyright (c) 2016 - Institut Pasteur
#
#  File author(s): Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: http://github.com/biokit
#
##############################################################################
from easydev import Logging, DevTools
from bioservices import IntactComplex
import pylab
import pandas as pd
from io import StringIO

__all__ = ['Complexes']


class Complexes(Logging):
    """Manipulate complexes of Proteins

    This class uses Intact Complex database to extract information about
    complexes of proteins. 

    When creating an instance, the default organism is "Homo sapiens". 
    The organism can be set to another one during the instanciation or later::

        >>> from biokit.network.complexes import Complexes
        >>> c = Complexes(organism='Homo sapiens')
        >>> c.organism = 'Rattus norvegicus'

    Valid organisms can be found in :attr:`organisms`. When changing the
    organism, a request to the Intact database is sent, which may take some
    time to update. Once done, information related to  this organism is stored
    in the :attr:`df` attribute, which is a Pandas dataframe. It 
    contains 4 columns. Here is for example one row::

        complexAC                                             EBI-2660609
        complexName                            COP9 signalosome variant 1
        description     Essential regulator of the ubiquitin (Ubl) con...
        organismName                                   Homo sapiens; 9606

    This is basic information but once a complex accession (e.g., EBI-2660609)
    is known, you can retrieve detailled information. This is done
    automatically for all the accession when needed. The first time, it will 
    take a while (20 seconds for 250 accession) but will be cache for this 
    instance. 

    The :attr:`complexes` contains all details about the entries found in
    :attr:`df`. It is a dictionary where keys are the complex accession. For 
    instance::

        >>> c.complexes['EBI-2660609']

    In general, one is interested in the participants of the complex, that is
    the proteins that form the complex. Another attribute is set for you::

        >>> c.participants['EBI-2660609']

    Finally, you may even want to obtain just the identifier of the participants
    for each complex. This is stored in the :attr:`identifiers`::

        >>> c.identifiers['EBI-2660609']

    Note however, that the identifiers are not neceseraly uniprot identifiers.
    Could be ChEBI or sometimes even set to None. The :meth:`strict_filter`
    removes the complexes with less than 2 (strictly) uniprot identifiers.
    
    Some basic statistics can be printed with :meth:`stats` that indeticates
    the number of complexes, number of identifiers in those complexes ,and
    number of unique identifiers. A histogram of number of appearance of each 
    identifier is also shown.

    The :meth:`hist_participants` shows the number of participants per complex.

    Finally, the meth:`search_complexes` can be used in the context of 
    logic modelling to infer the AND gates from a list of uniprot identifiers
    provided by the user. See :meth:`search_complexes` for details.

    Access to the Intact Complex database is performed using the 
    package BioServices provided in Pypi.
    """
    def __init__(self, organism='Homo sapiens', verbose=True, cache=False):
        """.. rubric:: Constructor

        :param str orgamism: the organism to look at. Homo sapiens
            is the default. Other possible organisms can be found
            in :attr:`organisms`.
        :param str verbose: a verbose level in ERROR/DEBUG/INFO/WARNING
            compatible with those used in BioServices.

        """
        super(Complexes, self).__init__(level=verbose)

        self.devtools = DevTools()
        self.webserv = IntactComplex(verbose=verbose, cache=cache)
        df = self.webserv.search('*', frmt='pandas')
        self.df = df

        #: list of valid organisms found in the database
        self.valid_organisms = list(set(df['organismName']))
        self.valid_organisms = [x.split(';')[0] for x in self.valid_organisms]


        #: list of valid organisms found in the database
        self.organisms = list(set(df['organismName']))
        self._organism = None
        if organism in self.organisms:
            self.organism = organism
        else:
            print("Organism not set yet. ")

        # This will populated on request as a cache/buffer
        self._details = None
        self._complexes = None

    def _get_organism(self):
        return self._organism
    def _set_organism(self, organism):
        self.devtools.check_param_in_list(organism, [str(x.split(";")[0]) for x in self.valid_organisms])
        self._organism = organism

        self.df = self.webserv.search('*', frmt='pandas',
                filters='species_f:("%s")' % self.organism)
        self._complexes = None
    organism = property(_get_organism, _set_organism, 
        doc="Getter/Setter of the organism")

    def hist_participants(self):
        """Histogram of the number of participants per complexes

        :return: a dictionary with complex identifiers as keys and
            number of participants as values

        ::

            from biokit.network.complexes import Complexes
            c = Complexes()
            c.hist_participants()

        """
        N = []
        count = {}
        for i, identifier in enumerate(self.complexes.keys()):
            n = len(self.complexes[identifier]['participants'])
            N.append(n)
            count[identifier] = n

        _ = pylab.hist(N, bins=range(0, max(N)))
        pylab.title('Number of participants per complex')
        pylab.grid()
        return count

    def stats(self):
        """Prints some stats about the number of complexes and histogram of the
        number of appearances of each species"""
        species = []
        for k in self.participants.keys():
            species.extend([x['identifier'] for x in self.participants[k]])
            N = []
        for spec in set(species):
            N.append(species.count(spec))
        _ = pylab.hist(N, bins=range(0, max(N)))
        pylab.title("Number of appaerances of each species")
        pylab.grid()
        print("""There are %s complexes involving %s participants with %s unique species. """ %
                 (len(self.complexes), len(species), len(set(species))))

    def _get_participants(self):
        participants = {}
        for k,v in self.complexes.items():
            participants[k] = v['participants']
        return participants
    participants = property(_get_participants, 
        doc="""Getter of the complex participants (full details)""")

    def _get_identifiers(self):
        identifiers = {}
        for k,v in self.participants.items():
            identifiers[k] = [x['identifier'] for x in v]
        return identifiers
    identifiers = property(_get_identifiers,
            doc="""Getter of the identifiers of the complex participants""")

    def _get_complexes(self):
        if self._complexes is None:
            self._load_complexes()
        return self._complexes.copy()
    complexes = property(_get_complexes,
        doc="""Getter of the complexes (full details""")

    def _load_complexes(self, show_progress=True):
        from easydev import Progress
        import time
        pb = Progress(len(self.df.complexAC))
        complexes = {}
        self.logging.info("Loading all details from the IntactComplex database")
        for i, identifier in enumerate(self.df.complexAC):
            res = self.webserv.details(identifier)
            complexes[identifier] = res
            if show_progress:
                pb.animate(i+1)
        self._complexes = complexes

    def remove_homodimers(self):
        """Remove identifiers that are None or starts with CHEBI
        and keep complexes that have at least 2 participants
        
        
        :return: list of complex identifiers that have been removed.
        """


        # None are actually homo dimers
        toremove = []
        for k,this in self.identifiers.items():
            remains = [x for x in this if x is not None]
            if len(remains)<=1:
                toremove.append(k)
        self.logging.info("removing %s homodimers complexes" % len(toremove))
        for this in toremove:
            del self._complexes[this]
        return toremove

    def search_complexes(self, user_species, verbose=False):
        """Given a list of uniprot identifiers, return complexes and
            possible complexes.

        :param list user_species: list of uniprot identifiers to be
            found in the complexes
        :return: two dictionaries. First one contains the complexes
            for which all participants have been found in the user_species
            list. The second one contains complexes for which some participants
            (not all) have been found in the user_species list.

        """
        level = self.debugLevel[:]
        if verbose:
            self.debugLevel = 'INFO'
        else:
            self.debugLevel = 'ERROR'

        and_gates = {}
        candidates = {}

        identifiers = self.identifiers.values()

        for k, identifiers in self.identifiers.items():

            # get rid of suffixes such as -1 or -PRO_xxx
            prefixes = [x.split("-")[0] if x is not None else x for x in identifiers]


            # You may have a complex with ['P12222', 'P33333-PRO1',
            # 'P33333-PRO2'], in which case P33333 is found only once and
            # thereofre the final number of found participants is not the length
            # of the complexes...so we need to get rid of the duplicates if any
            prefixes = list(set(prefixes))
            N = len(prefixes)
            found = [spec for spec in user_species if spec in prefixes]

            if len(found) == N:
                self.logging.info('Found entire complex %s ' % k)
                and_gates[k] = identifiers[:]
            elif len(found) >= 1:
                self.logging.info('Found partial complex %s with %s participants out of %s' % 
                        (k, len(found), len(identifiers)))
                candidates[k] = {'participants': identifiers, 'found': found}
        self.debugLevel = level[:]
        return and_gates, candidates

    def search(self, name):
        """Search for a unique identifier (e.g. uniprot) in all complexes

        :return: list of complex identifiers where the name was found
        """
        found = []
        for k, identifiers in self.identifiers.items():
            prefixes = [x.split("-")[0] if x is not None else x for x in identifiers ]
            if name in prefixes:
                self.logging.info("Found %s in complex %s (%s)" % (name, k,
                    identifiers))
                found.append(k)
        return found

    def chebi2name(self, name):
        """Return the ASCII name of a CHEBI identifier"""
        from bioservices import ChEBI
        c = ChEBI()
        name = dict(c.getLiteEntity(name)[0])['chebiAsciiName']
        return name

    def uniprot2genename(self, name):
        """Return the gene names of a UniProt identifier"""
        from bioservices import UniProt
        c = UniProt(cache=True)

        try:
            res = pd.read_csv(StringIO(c.search(name, limit=1)), sep='\t')
            return list(res['Gene names'].values)
        except:
            print("Could not find %s" % name)

    def report(self, species):
        complete, partial = self.search_complexes(species, verbose=False)
        res = {'Found':[], 'Participants':[], 'Complete':[],
                'Identifier':[], 'Number found':[], 'Number of participants':[],
                'Name':[]}

        for k, v in complete.items():
            res['Name'].append(self.complexes[k]['name'])
            res['Found'].append(";".join(v))
            res['Number found'].append(len(v))
            res['Participants'].append(";".join(self.identifiers[k]))
            res['Number of participants'].append(len(self.identifiers[k]))
            res['Complete'].append(True)
            res['Identifier'].append(k)

        for k, v in partial.items():
            res['Name'].append(self.complexes[k]['name'])
            res['Found'].append(";".join(v['found']))
            res['Number found'].append(len(v['found']))
            res['Participants'].append(";".join(self.identifiers[k]))
            res['Number of participants'].append(len(self.identifiers[k]))
            res['Complete'].append(False)
            res['Identifier'].append(k)

        res = pd.DataFrame(res, columns=['Found', 'Participants', 'Identifier', 'Name', 'Number found', 
            'Number of participants', 'Complete'])
        return res













# -*- python -*-
#
#  This file is part of biokit software
#
#  Copyright (c) 2014-
#
#  File author(s): Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/biokit
#
##############################################################################
from .session import RSession

__all__ = ["bool2R", "rcode"]


def bool2R(value):
    """Transforms a boolean into a R boolean value T or F"""
    if value is True:
        return "T"
    if value is False:
        return "F"
    else:
        raise ValueError("expecting a boolean value")


def rcode(code, verbose=True):
    """Run a R script and return the RSession object"""
    r = RSession(verbose=verbose)
    r.run(code)
    return r







import tempfile
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
except:
    from urllib2 import urlopen, HTTPError, URLError


import os.path

from biokit.rtools import bool2R, RSession
from distutils.version import StrictVersion
from easydev import Logging, TempFile


__all__ = ["get_R_version", "biocLite", "RPackage", 
    'install_package', 'RPackageManager']


def install_package(query, dependencies=False, verbose=True,
    repos = "http://cran.univ-paris1.fr/"):
    """Install a R package

    :param str query: It can be a valid URL to a R package (tar ball), a CRAN
        package, a path to a R package (tar ball), or simply the directory
        containing a R package source.
    :param bool dependencies:
    :param repos: if provided, install_packages automatically select the
        provided repositories otherwise a popup window will ask you to select a repo

    ::

        >>> rtools.install_package("path_to_a_valid_Rpackage.tar.gz")
        >>> rtools.install_package("http://URL_to_a_valid_Rpackage.tar.gz")
        >>> rtools.install_package("hash") # a CRAN package
        >>> rtools.install_package("path to a valid R package directory")

    .. seealso:: :class:`biokit.rtools.RPackageManager`
    """
    session = RSession(verbose=verbose)

    # Is it a local file?
    if os.path.exists(query):
        repos = 'NULL'
    else:
        repos = '"{0}"'.format(repos) # we want the " to be part of the string later on
    
    try:
        # PART for fetching a file on the web, download and install locally
        if verbose:
            print("Trying from the web ?")
        data = urlopen(query)
        fh = TempFile(suffix=".tar.gz")
        with open(fh.name, 'w') as fh:
            for x in data.readlines():
                fh.write(x)
        code = """install.packages("%s", dependencies=%s """ % \
            (fh.name, bool2R(dependencies))
        code += """ , repos=NULL) """
        session.run(code)

    except Exception as err:
        if verbose:
            print(err)
            print("trying local or from repos")
            print("RTOOLS warning: URL provided does not seem to exist %s. Trying from CRAN" % query)
        code = """install.packages("%s", dependencies=%s """ % \
            (query, bool2R(dependencies))

        code += """ , repos=%s) """ % repos
        session.run(code)
        return

def get_R_version():
    """Return R version"""
    r = RSession()
    r.run("version")
    return r.version


def biocLite(package=None, suppressUpdates=True, verbose=True):
    """Install a bioconductor package

    This function does not work like the R function. Only a few options are
    implemented so far. However, you can use rcode function directly if needed.

    :param str package: name of the bioconductor package to install. If None, no
        package is installed but installed packages are updated. If not provided, 
        biocLite itself may be updated if needed.
    :param bool suppressUpdates: updates the dependencies if needed (default is
        False)

    :return: True if update is required or the required package is installed and
        could be imported. False otherwise.

    ::

        >>> from biokit.viz.rtools import biocLite
        >>> biocLite("CellNOptR")

    """
    code = """source("http://bioconductor.org/biocLite.R")\n"""

    # without a package, biocLite performs an update of the installed packages
    if package is None:
        code += """biocLite(suppressUpdates=%s) """ % (
            bool2R(suppressUpdates))
    else:
        # if not found, no error is returned...
        code += """biocLite("%s", suppressUpdates=%s) """ % (
            package,
            bool2R(suppressUpdates))
    r = RSession(verbose=verbose)
    r.run(code)
    

class RPackage(object):
    """

    ::

        >>> from biokit.rtools import package
        >>> p = package.RPackage('CellNOptR')
        >>> p.isinstalled
        True
        >>> p.version
        '1.11.3'

    .. todo:: do we need the version_required attribute/parameter anywhere ?

    .. note:: R version includes dashes, which are not recognised
       by distutils so they should be replaced. 
    """
    def __init__(self, name, version_required=None, install=False, verbose=False):
        self.name = name
        self.version_required = version_required  # the required version
        if self.version_required and isinstance(self.version_required, str) is False:
            raise TypeError("version_required argument must be a string e.g., 2.0, 2.0.1")
        if version_required and "." not in self.version_required:
            # trying to infer correct version
            self.version_required += '.0'

        self.session = RSession()

        code = """rvar_version = as.character(packageVersion("%s"))"""
        self.session.run(code  % (self.name))
        try:
            self._version = self.session.rvar_version
        except:
            self._version = None

        if self.version is None and install is True:
            self.install(name)
        if self.version and self.version_required:
            if self._get_val_version(self.version) >= self._get_val_version(self.version_required):
                pass
            else:
                print("Found %s (version %s) but version %s required." % (
                    self.name, self.version, self.version_required))

    def _get_val_version(self, version):
        return StrictVersion(version.replace("-", "a"))

    def install(self):
        install_package(self.name)

    def _get_isinstalled(self):
        if self.version:
            return True
        else:
            return False
    isinstalled = property(_get_isinstalled)

    def _get_version(self):
        return self._version
    version = property(_get_version)

    def __str__(self):
        if self.version:
            txt = self.name + ": " + self.version
        else:
            txt = self.name
        return txt


class RPackageManager(object):
    """Implements a R package manager from Python

    So far you can install a package (from source, or CRAN, or biocLite)

    ::

        pm = PackageManager()
        [(x, pm.installed[x][2]) for x in pm.installed.keys()]


    You can access to all information within a dataframe called **packages** where
    indices are the name packages. Some aliases are provided as attributes (e.g., available, 
    installed)


    """
    cran_repos = "http://cran.univ-lyon1.fr/"

    def __init__(self, verbose=True):
        self.session = RSession()
        self.logging = Logging("INFO")
        self.logging.info('Fetching package information')
        self.update()

    def _update(self):
        # local import ?
        import numpy
        import pandas
        # figure out the installed packages first
        code = """rvar_packages = as.data.frame(installed.packages())"""
        self.session.run(code)
        s = self.session.rvar_packages
        # FIXME. these 4 lines are needed as a hack related to pyper.
        try:
            s = s.replace("\n", "")
            df = eval(s)
        except:
            df = s

        df.set_index('Package', inplace=True)
        self._packages = df.copy()

        # Now, fetch was is possible to install from the default cran repo
        code = """rvar_status=packageStatus(repos="%s/src/contrib")"""
        code = code % self.cran_repos

        self.session.run(code)
        s = self.session.rvar_status

        # FIXME.
        try:
            s = s.replace("\n", "")
            res = eval(s)
        except:
            res = s
        res['inst'].set_index('Package', inplace=True)
        res['avail'].set_index('Package', inplace=True)
        self._status = res

    def update(self):
        """If you install/remove packages yourself elsewhere, you may need to 
        call this function to update the package manager"""
        try:
            #self.session.reconnect()          
            self._update()
        except:
            self.logging.warning("Could not update the packages. Call update() again")
        
    def _compat_version(self, version):
        return version.replace("-", "a")

    def _get_installed(self):
        # we do not buffer because packages may be removed manually or from R of
        # using remove_packages method, ....
        #self._package_status()
        return self._status['inst']
    installed = property(_get_installed, "returns list of packages installed as a dataframe")

    def _get_available(self):
        # we do not buffer because packages may be removed manually or from R of
        # using remove_packages method, ....
        #self._package_status()
        return self._status['avail']
    available = property(_get_available, "returns list of packages available as a dataframe")

    def  _get_packages(self):
        # do not buffer since it may change in many places
        return self._packages
    packages = property(_get_packages)

    def get_package_latest_version(self, package):
        """Get latest version available of a package"""
        return self.available['Version'].ix[package]

    def get_package_version(self, package):
        """Get version of an install package"""
        if package not in self.installed.index:
            self.logging.error("package {0} not installed".format(package))
        return self.installed['Version'].ix[package]

    def biocLite(self, package=None, suppressUpdates=True, verbose=False):
        """Installs one or more biocLite packages

        :param package: a package name (string) or list of package names (list of 
            strings) that will be installed from BioConductor. If package is set 
            to None, all packages already installed will be updated.

        """
        if isinstance(package, str):
            if package not in self.installed.index:
                biocLite(package, suppressUpdates, verbose=verbose)
        elif isinstance(package, list):
            for pkg in package:
                self.logging.info("Installing %s" % pkg)
                if self.is_installed(pkg) is False:
                    biocLite(pkg, suppressUpdates, verbose=verbose)
        else: # trying other cases (e.g., None updates biocLite itself). 
            biocLite(package, suppressUpdates, verbose=verbose)
        self.update()

    def _isLocal(self, pkg):
        if os.path.exists(pkg):
            return True
        else:
            return False

    def remove(self, package):
        """Remove a package (or list) from local repository"""
        rcode ="""remove.packages("%s")"""
        if isinstance(package, str):
            package = [package]
        for pkg in package:
            if pkg in self.installed.index:
                self.session(rcode % pkg)
            else:
                self.logging.warning("Package not found. Nothing to remove")
        self.update()

    def require(self, pkg, version):
        "Check if a package with given version is available"

        if pkg not in self.installed.index:
            self.logging.info("Package %s not installed" % pkg)
            return False
        currentVersion = self.packageVersion(pkg)
        if self._get_version(currentVersion) >= self._get_version(version):
            return True
        else:
            return False

    def _install_package(self, packageName, dependencies=True):
        """Installs one or more CRAN packages
        
        .. todo:: check if it is already available to prevent renstallation ?
        """

        repos = self.cran_repos
        # if this is a source file we want to reset the repo
        if isinstance(packageName, str):
            packageName = [packageName]
        for pkg in packageName:
            if self.is_installed(pkg) is False:
                self.logging.info("Package not found. Installing %s..." % pkg)
                install_package(pkg, dependencies=dependencies, 
                        repos=repos)
            else:
                self.logging.info("Package %s found. " % pkg)
                install_package(pkg, dependencies=dependencies, 
                        repos=repos)
        self.update()

    def install(self, pkg, require=None, update=True, reinstall=False):
        """install a package automatically scanning CRAN and biocLite repos

        if require is not set and update is True, when a newest version of a package
        is available, it is installed

        """
        from easydev import to_list
        pkgs = to_list(pkg)
        for pkg in pkgs:
            self._install(pkg, require=require, update=update, reinstall=reinstall)

    def _install(self, pkg, require=None, update=update, reinstall=False):
        # LOCAL file
        if self._isLocal(pkg):
            # if a local file, we do not want to jump to biocLite or CRAN. Let
            # us install it directly. We cannot check version yet so we will
            # overwrite what is already installed
            self.logging.warning("Installing from source")
            self._install_package(pkg)
            return

        # From CRAN
        if self.is_installed(pkg):
            currentVersion = self.get_package_version(pkg)
            # if not provided, require should be the latest version
            if require is None and update is True:
                try:
                    require = self.get_package_latest_version(pkg)
                except:
                    # a non-cran package (bioclite maybe)
                    pass

            if require is None:
                self.logging.info("%s already installed with version %s" % \
                    (pkg, currentVersion))
                return
            
            # if require is not none, is it the required version ?
            if self._get_version(currentVersion) >= self._get_version(require) and reinstall is False:
                self.logging.info("%s already installed with required version %s" \
                    % (pkg, currentVersion))
                # if so, nothing to do
            else:
                # Try updating
                self.logging.info("Updating")
                self._install_package(pkg)
                if require is None:
                    return
                currentVersion = self.get_package_version(pkg)
                if self._get_version(currentVersion) < self._get_version(require):
                    self.logging.warning("%s installed but current version (%s) does not fulfill your requirement" % \
                        (pkg, currentVersion))

        elif pkg in self.available.index:
            self._install_package(pkg)
        else:
            # maybe a biocLite package ?
            # require is ignored. The latest will be installed
            self.logging.info("Trying to find the package on bioconductor")
            self.biocLite(pkg)
            if require is None:
                return
            currentVersion = self.get_package_version(pkg)
            if self._get_version(currentVersion) >= self._get_version(require):
                self.logging.warning("%s installed but version is %s too small (even after update)" % \
                    (pkg, currentVersion, require))

    def _get_version(self, version):
        # some pacakge do not use the correct version convention
        try:
            return StrictVersion(version)
        except:
            try:
                return StrictVersion(version.replace("-", "a"))
            except:
                # snowfall package example was 1.86-6.1
                # This becomes 1.86a61  which is not great but not workaround
                # for now
                left, right = version.split("-")
                version = left + "a" + right.replace('.', '')
                return StrictVersion(version)

    def is_installed(self, pkg_name):
        if pkg_name in self.installed.index:
            return True
        else:
            return False





"""Plotting tools


.. autosummary::

    biokit.viz.corrplot.Corrplot
    biokit.viz.hinton.hinton
    biokit.viz.hist2d.Hist2D
    biokit.viz.imshow.Imshow
    biokit.viz.scatter.ScatterHist
    biokit.viz.volcano.Volcano
    biokit.viz.heatmap.Heatmap

"""

from .corrplot import *
from .hinton import *
from .hist2d import *
from .imshow import *
from .scatter import *
from .volcano import *
from .heatmap import *



"""GO terms


Resources: geneontology.org and godb Python package from endrebak user (Bakken).
https://github.com/endrebak/godb/

"""
import urllib
import os
from collections import defaultdict

from biokit import biokitPATH
import pandas as pd


__all__ = ['num2goid', 'GOId', 'GODB', 'GOTerm']


def num2goid(value):
    """Returns GO identifier string correctly formatted"""
    return GOId(value).identifier


class GOId(object):
    """Representation of a GO identifier

    A GO identifier simply takes the form GO:XXXXXXX where X are digits
    so this is the string "GO:" followed by 7 digits.

    This class ease the construction and validation of GO identifiers.
    As an input, one can provide a valid GO identifiers or a number, which
    is then transformed into the proper formatted identifier

    ::

        >>> o = GOId(5)
        >>> o.iddentifier
        'GO:0000005'
        >>> o = GOId('GO:0000005')
        >>> o.identifier
        'GO:0000005'
        >>> print(o)
        'GO:0000005'

    """
    def __init__(self, value):
        """converts the numbers into Gene Ontology IDs.

        :param list numbers: a list of integer or strings or just a
            string or int. The numbers must be
        :return: list of GO Ids

        IDs are seven-digit numbers preceded by the prefix GO: (Gene Ontology
        database standard).
        """
        self._identifier = None
        self.identifier = value

    def _get_identifier(self):
        return self._identifier
    def _set_identifier(self, identifier):
        self._identifier = self._num2goid(identifier)
    identifier = property(_get_identifier, _set_identifier)

    def _num2goid(self, value):
        msg = "GO identifier format must be GO:XXXXXXX with X being numbers"

        if isinstance(value, str):
            if value.startswith("GO:"):
                prefix, suffix = value.split(":", 1)
            else:
                suffix = value
            suffix = int(suffix)
        elif isinstance(value, int):
            suffix = value
        else:
            raise TypeError("value must be a string or an integer. Provided %s"
                    % value)

        if suffix < 1e7:
            goId = "GO:%07d" % suffix
        else:
            raise ValueError(msg)
        return goId

    def __repr__(self):
        return self.identifier

    def __str__(self):
        return self.identifier


class GOTerm(object):
    """Transform a text-version of a GO Term into data structures

    A text contain information about a GO term looks like (obo format)::

        [Term]
        id: GO:0000003
        name: reproduction
        namespace: biological_process

    This can be parsed with this class. You can retrieve these OBO text with
    other tools from BioServices::

        from bioservices import QuickGO
        q = QuickGO()
        data = q.Term("GO:0003824", frmt='obo')

        from biokit import GOTerm
        term = GOTerm(data)
        term.to_dict()

    If an item appears several times (e.g., xref), then the values stored is a
    list, otherwise, it is the raw text found after the item name.

    Here is a brief overview of the structure of an ontology (see
    geneontology.org for details).

    Unique identifier **id** and term name: every term (e.g., mitochondrion) has
    a unique zero-padded seven digit identifier (often called the
    term accession or term accession number) prefixed by GO:, e.g. GO:0005125.

    Namespace (**namespace**) denotes which of the three
    sub-ontologies-cellular component, biological process or molecular
    function-the term belongs to.

    Definition: a textual description of what the term represents, plus
    reference(s) to the source of the information.

    Relationships to other terms: One or more links that capture how the term
    relates to other terms in the ontology. All terms have an "is a" sub-class
    relationship to another term. Gene Ontology employs a number of other
    relations, including 'part of' , and 'regulates'.

    Synonyms **synonym** are alternative words or phrases closely related in
    meaning to the term name, with indication of the relationship between the
    name and synonym given by the synonym scope. The scopes for GO synonyms are
    EXACT, BROAD, NARROW and RELATED.

    Database cross-references or dbxrefs, refer to identical or very similar
    objects in other databases. For instance, the molecular function term
    retinal isomerase activity is cross-referenced with the Enzyme Commission
    entry EC:5.2.1.3; the biological process term sulfate assimilation has the
    cross-reference MetaCyc:PWY-781.

    Comment: An y extra information about the term and its usage.

    Subset (*) : Indicates that the term belongs to a designated subset
        of terms, e.g. one of the GO slims.

    Obsolet tag: Indicates that the term has been deprecated and should not be
        used.

    **id** and **name** are compulsary and unique.
    optional tags are **is_

    .. seealso:: :class:`GODB` and QuickGO from bioservices. For instance,
        from QuickGO, you may get more information about cross references as
        compared to :class:`GODB` that relies on geneontologies.org snapshot.
    """
    def __init__(self, text, remove_comments=True):
        if text.startswith("<obo>"):
            raise NotImplementedError("obo XML format, use obo plain text instead")
        elif "id:" in text and "name:" in text:
            #assume that the text is a OBO formatted text
            self.text = text[:]
        else:
            # Assume the input text is a valid GO identifier
            goid = GOId(text).identifier
            # try to retrieve Term for this GO identifier with QuickGO
            from bioservices import QuickGO
            q = QuickGO()
            text = q.Term(goid, frmt='obo')
            self.text = text
        self.remove_comments = remove_comments

    def _remove_comments(self, term):
        # Let us get rid of all comments, this will make further parsing easier
        for k, v in term.items():
            if isinstance(v, list):
                term[k] = [x.split("!")[0].strip() for x in v]
            else:
                if "!" in v:
                    term[k] = v.split("!")[0].strip()
        return term

    def to_dict(self):
        # We assume all tags are list, which is not true
        d = defaultdict(list)
        for this in self.text.strip().split("\n"):
            if ":" in this:
                key, value = this.split(":", 1)
                d[key.strip()].append(value.strip())

        # Based on http://oboformat.googlecode.com/svn/trunk/doc/GO.format.obo-1_2.html
        # we can clean up the lists converting some of them to
        # single item. We also issue a warning with deprecated ones.
        lists = ['alt_id', 'synonym', 'is_a', 'xref', 'intersection_of',
            'relationship', 'union_of', 'disjoint_from', 'consider',
            'replaced_by', 'subset', 'property_value']
        nonlist = ['id', 'name', 'is_anonymous', 'is_obsolete', 'def',
                'comment', 'created_by', 'creation_date', 'namespace']
        deprecated = {
                'exact_synonym': 'synonym',
                'narrow_synonym':'synonym',
                'broad_synonym': 'synonym',
                'xref_analog': 'xref',
                'xref_unk': 'xref',
                'use_term': 'consider'
                }
        # missing: namespace, subset,
        dd = {}
        for k, v in d.items():
            if k in lists: # nothing to do
                dd[k] = v
            elif k in nonlist:
                if len(v) == 1:
                    dd[k] = v[0]
                else:
                    raise ValueError("%s must be found only once. Check %s" %
                            (k, d['id']))
            elif k in deprecated.keys():
                print("%s deprecated. Kept and assuming non unique tag" %
                        (k, d['id']))
                dd[k] = v
            else:
                print("%s not handled in %s. Assuming non unique tag" %
                    (k, d['id']))
                dd[k] = v
        if self.remove_comments is True:
            dd = self._remove_comments(dd)

        # IN OBO/GO format, id and name must be provided
        # others are optional
        if 'id' not in dd.keys() and 'name' not in dd.keys():
            raise ValueError("'id' tag and 'name' must be provided")

        return dd


class GODB(object):
    """Simple handler to get list of GO terms from geneontology website


    The Gene Ontology project provides controlled vocabularies of defined terms
    epresenting gene product properties. These cover three domains: Cellular
    omponent, the parts of a cell or its extracellular environment; Molecular
    unction, the elemental activities of a gene product at the molecular level,
    uch as binding or catalysis; and Biological Process, operations or sets of
    olecular events with a defined beginning and end, pertinent to the
    functioning of integrated living units: cells, tissues, organs, and
    organisms.


    .. seealso:: bioservices QuickGO class
    """
    def __init__(self, name='go.obo', drop_obsolet=True, local=False):
        """

        Searches for a file on geneontology.org exce

        :param name: name of the go OBO file to be downloaded
        :param drop_obsolet: drop obsolet GO terms from the entire DB
        :param local: read the OBO file locally (False) or downloads
           from geneontology.org if not present in the biokit
           directory.
        """
        self.name = 'go.obo'
        if local is False:
            self.filename = biokitPATH + os.sep + self.name
        else:
            self.filename = self.name
        self._init()

        if drop_obsolet is True:
           self.df = self.df[self.df.is_obsolete != True]
           self.df.pop('is_obsolete')

        # split relationships

        self.mapping = {'id': 'GO id',
                'namespace': 'Ontology',
                'name': 'Term',
                'def': 'Definition'}
        # TODO: split relationshio into has_part and part_of

    def _init(self):
        if os.path.exists(biokitPATH + os.sep + self.name) is False:
            self._download_godb()
        # Read the original OBO file
        self.read_goterms()
        # Parse the GO terms one by one and transform into a DataFrame
        self.df = self._terms2df()

    def _download_godb(self):
        print("Downloading go db once for all")
        urllib.urlretrieve ("http://geneontology.org/ontology/%s" % self.name,
             self.filename)

    def read_goterms(self):
        # lines starting with ! can be ignored
        # any line may end with a comment starting with !
        with open(self.filename) as fhdata:
            data = "".join(fhdata.readlines()).split("\n\n[Term]")[1:-1]
            self.obo_terms = data

    def _term2dict(self, term):
        term = GOTerm(term)
        return term.to_dict()

    def _terms2df(self):
        terms = [self._term2dict(term) for term in self.obo_terms]
        df = pd.DataFrame(terms)
        df.replace('true', True, inplace=True)
        df.replace('false', False, inplace=True)
        df.set_index('id', inplace=True)
        return df

    def __len__(self):
        return len(self.df)

    def get_annotations(self):
        # replace some columns
        df = self.df.copy()
        df = df[['namespace', 'name', 'synonym', 'def', 'relationship']]
        df.columns = ["Ontology", "Term", "Synonym", "Definition",
                "relationship"]
        return df

    def search(self, search, where='name', method='in'):
        if method == 'in':
            selection = self.df[where].apply(lambda x: search in x)
        elif method == 'is':
            selection = self.df[where].apply(lambda x: search == x)
        elif method == 'startswith':
            selection = self.df[where].apply(lambda x: x.startswith(search))
        return self.df[selection].copy()

    #  The is a relation is transitive, which means that if A is a B, and B is a
    #  C, we can infer that A is a C.
    #  B part of A does not mean that A have B systematically (think of
    # replication fork part of chromosome but chromosome do not have a
    # replication fork systemtically.
    #  Like is a, part of is transitive: if A part of B part of C
    #  then A part of C
    #
    # The logical complement to the part of relation is has part, which
    # represents a part-whole relationship from the perspective of the parent.
    # As with part of, the GO relation has part is only used in cases where A
    # always has B as a part, i.e. where A necessarily has part B. If A exists,
    # B will always exist; however, if B exists, we cannot say for certain that
    # A exists. i.e. all A have part B; some B part of A.
    def get_children(self, ontology='CC',
            relations=['is_a', 'part_of', 'has_part']):
        """http://geneontology.org/page/ontology-relations"""

        if ontology in ['CC', 'cellular_component']:
            ontology = 'cellular_component'
        elif ontology in ['BP', 'biological_process']:
            ontology = 'biological_process'
        elif ontology in ['MF', 'molecular_function']:
            ontology = 'molecular_function'
        else:
            raise ValueError("ontology must be one of %s" % set(self.df.namespace))

        newdf = self.df.loc[self.df.namespace == ontology]

        children = []
        parents = []
        vrelations = []
        # this part is slow and maybe improved.
        for name, row in newdf.iterrows():
            if str(row.is_a) == 'nan':
                pass
            elif 'is_a' in relations:
                for this in row.is_a:
                    children.append(name)
                    parents.append(this)
                    vrelations.append('is_a')
            if str(row.relationship) == 'nan':
                pass
            else:
                for this in row.relationship:
                    relation, goid = this.split()
                    if relation == 'part_of' and relation in relations:
                        children.append(name)
                        parents.append(goid)
                        vrelations.append(relation)
                    elif relation == 'has_part' and relation in relations:
                        # not that children/parent are swapped as compared
                        # to the part_of relation.
                        children.append(goid)
                        parents.append(name)
                        vrelations.append(relation)

        df = pd.DataFrame({'Child':children, 'Parent':parents,
            'Relation':vrelations})
        df.sort('Relation', inplace=True, ascending=False)
        return df

    def get_offspring(self):
        # again, based on
        # https://github.com/endrebak/godb/
        # slower but no need for now
        offspring_df = self.get_children().copy()[["Child", "Parent"]]
        offspring_df.columns = ["Offspring", "Parent"]

        df = offspring_df
        new_df = self._compute_transitive_closure(df)
        while not df.equals(new_df):
            df = df.append(new_df).drop_duplicates()
            new_df = self._compute_transitive_closure(df)

            df = df.drop_duplicates().sort()
            new_df = new_df.drop_duplicates().sort()
        return new_df

    def _compute_transitive_closure(self, df):
        """Computes the transitive closure from a two-col parent/child map."""
        df_temp = df[df["Parent"].isin(df["Offspring"])]
        df2 = df.merge(df_temp, left_on="Offspring", right_on="Parent",
            suffixes=["_1_gen", "_2_gen"])
        df2 = df2.drop(["Offspring_1_gen", "Parent_2_gen"], axis=1)
        df2.columns = ["Parent", "Offspring"]
        concat_df = pd.concat([df, df2]).drop_duplicates()
        return concat_df



"""Core function for the plotting tools"""
import pandas as pd


__all__ = ["VizInput2D"]


class VizInputSquare(object):
    def __init__(self, x, verbose=False):
        self.verbose = verbose
        self.df = pd.DataFrame(x)


class VizInput2D(object):
    def __init__(self, x, y=None, verbose=False):
        self.verbose = verbose

        self.xy_names = ['x', 'y']
        if isinstance(x, pd.DataFrame) is True:
            self.df = x.copy()
            columns = list(self.df.columns)
            columns[0] = 'x'
            columns[1] = 'y'
            self.xy_names = self.df.columns[0:2]
            self.df.columns = columns
        elif y is None:
            # could be a list of lists, a pandas-compatible dictionary
            self.df = pd.DataFrame(x)
            if self.df.shape[1] != 2:
                if self.df.shape[0] == 2:
                    print("warning transposing data")
                    self.df = self.df.transpose()
        elif x is not None and y is not None:
            self.df = pd.DataFrame({'x':x, 'y':y})
        else:
            raise ValueError('not a dataframe or list of items or dictionary.')







"""Imshow utility"""
from biokit.viz.core import VizInput2D
import pylab
import pandas as pd


__all__ = ['imshow', 'Imshow']


class Imshow(VizInput2D):
    def __init__(self, x, y=None, verbose=True):
        super(Imshow, self).__init__(x=x, y=y, verbose=verbose)

    
    def imshow(interpolation='None', aspect='auto', cmap='hot', tight_layout=True,
        colorbar=True, fontsize_x=None, fontsize_y=None, rotation_x=90,
        xticks_on=True, yticks_on=True, **kargs):
        """wrapper around imshow to plot a dataframe


        if list of lists or numpy array, no information about x and y axis will be filled
        otherwise, uses indedx and columns from the dataframe.
        """

        if isinstance(data, pd.DataFrame):
            pass
        else:
            # try a cast to df
            data = pd.DataFrame(data)

        pylab.clf()
        pylab.imshow(data, interpolation=interpolation, aspect=aspect, cmap=cmap, **kargs)

        if fontsize_x == None:
            fontsize_x = 16 #FIXME use default values
        if fontsize_y == None:
            fontsize_y = 16 #FIXME use default values

        if yticks_on is True:
            pylab.yticks(range(0, len(data.index)), data.index, 
                fontsize=fontsize_y)
        else:
            pylab.yticks([])
        if xticks_on is True:
            pylab.xticks(range(0, len(data.columns[:])), data.columns, 
                fontsize=fontsize_x, rotation=rotation_x)
        else:
            pylab.xticks([])

        if colorbar is True:
            pylab.colorbar()

        if tight_layout:
            pylab.tight_layout()








"""

resources: http://en.wikipedia.org/wiki/FASTQ_format

FASTQ format supported are those from Sanger and version CASAVA 1.8 (Illumina).
You may use the version <1.4 and identifiers would be recognised but we do not
support any export of other features except for casava>=1.8

The original Sanger PHRED standard is used for quality.


resources: https://github.com/agordon/fastx_toolkit

"""
import copy
import math
import pandas as pd
import pylab
from easydev import AttrDict

class FASTQIdentifier(object):
    """Container for the FASTQ identifier"""
    def __init__(self, identifier):
        self._identifier = identifier
        if self.identifier.startswith("@") is False:
            raise ValueError("identifier must start with @ character")        
    def _get_identifier(self):
        return self._identifier
    identifier = property(_get_identifier)


class FASTQIdentifierIllumnina(FASTQIdentifier):
    """Container for the FASTQ illumina container

    This is an old illumia version:: 
    
        @HWUSI-EAS100R:6:73:941:1973#0/1

    =============== =================================================================
    =============== =================================================================
    HWUSI-EAS100R   the unique instrument name
    6               flowcell lane
    73              tile number within the flowcell lane
    941             'x'-coordinate of the cluster within the tile
    1973            'y'-coordinate of the cluster within the tile
    #0              index number for a multiplexed sample (0 for 
                    no indexing)
    /1              the member of a pair, /1 or /2 (paired-end or 
                    mate-pair reads only)
    =============== =================================================================

    This is the new supported format::

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

    =============== ===========================================================
    =============== ===========================================================
    EAS139          the unique instrument name
    136             the run id
    FC706VJ         the flowcell id
    2               flowcell lane
    2104            tile number within the flowcell lane
    15343           'x'-coordinate of the cluster within the tile
    197393          'y'-coordinate of the cluster within the tile
    1               the member of a pair, 1 or 2 (paired-end or mate-pair 
                    reads only)
    Y               Y if the read is filtered, N otherwise
    18              0 when none of the control bits are on, otherwise it 
                    is an even number
    ATCACG          index sequence
    =============== ===========================================================

    """
    def __init__(self, identifier):
        super(FASTQIdentifierIllumnina, self).__init__(identifier)
        if "#" in self.identifier:
            self._interpret_identifier_1_4()
        else:
            self._interpret_identifier()

    def _interpret_identifier(self):
        """

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

        Note the space and : separators
        """
        # skip @ character 
        identifier = self.identifier[1:]
        # replace spaces by : character
        identifier = ' '.join(identifier.split())
        identifier = identifier.replace(' ', ':')
        items = identifier.split(':')
        if len(items) != 11:
            raise ValueError('Number of items in the identifier should be 11')
        self._instrument_name = items[0]
        self._run_id = items[1]
        self._flowcell_id = items[2]
        self._flowcell_lane = items[3]
        self._tile_number = items[4]
        self._x_coordinate = items[5]
        self._y_coordinate = items[6]
        self._member_pair = items[7]
        self._filtered = items[8]
        self._control_bits = items[9]
        self._index_sequence = items[10]

    def _interpret_identifier_1_4(self):
        # skip @ character 
        identifier = self.identifier[1:]
        identifier = identifier.replace('#', ':')
        identifier = identifier.replace('/', ':')
        items = identifier.split(':')
        
        if len(items) != 7:
            raise ValueError('Number of items in the identifier should be 11')
        # ['@HWUSI-EAS100R', '6', '73', '941', '1973#0/1']
        self._instrument_name = items[0]
        self._flowcell_lane = items[1]
        self._tile_number = items[2]
        self._x_coordinate = items[3]
        self._y_coordinate = items[4]
        self._index = '#' + items[5]
        self._member_pair = '/' + items[6]

    def _get_instrument_name(self):
        return self._instrument_name
    instrument_name = property(_get_instrument_name)



class FASTQIdentifierNCBI(FASTQIdentifier):
    """Container for the NCBI FASTQ identifier"""
    def __init__(self, identifier):
        super(FASTQNCBI, self).__init__(identifier)



class SingleFASTQ(object):
    _quality_character = "".join([chr(x) for x in range(33, 33+94)])
    def __init__(self, data=None):
        self.identifier = None
        self.sequence = None
        self.quality = None
        
        # Illumina 1.3 to 1.7 style FASTQ file using PHREDscores with offset 64
        self.offset = 33

        if data is not None:
            if isinstance(data, dict):
                print('init with dict')
                fastq = self.from_dict(data)
                self.sequence = fastq.sequence[:]
                self.quality = fastq.quality[:]
                self.identifier = fastq.identifier[:]
            else:
                self.read(data)

    def read(self, data):
        # could be a file or a string. A filename is also a string
        # so isinstance won't work. 
        try:
            with open(data, "r") as fh:
                data = fh.read()
            self.data = data[:]
        except IOError:
            self.data = data[:]
        self._parse_data(data)        

    def _parse_data(self, data):
        self._init()
        for i, line in enumerate(data.split("\n")):
            # skip blankline
            if len(line) == 0:
                continue
            # have we finished to parse the sequence ?
            # if + is found in the new line, then the answer is yes
            if line.startswith("+") and self._parsing_mode == 'sequence': 
                self._parsing_mode = 'quality'
                continue

            # assume identifier is on 1 line only
            if self._parsing_mode == 'identifier':
                if line.startswith("@"):
                    self.identifier = FASTQIdentifier(line) # check
                    self.identifier = self.identifier.identifier
                    self._parsing_mode = 'sequence'
                else:
                    raise ValueError("Expected @ at the beginning of the line %s" % i)
            elif self._parsing_mode == 'sequence':
                self.sequence += line
                self.nolines += 1 
            elif self._parsing_mode == 'quality':
                self.quality += line
                # here the line may start with @, which is confusing with the identifier
                # however, from the sequence we know how many lines are expected
                self.nolines -= 1
                if self.nolines == 0:
                    # clean sequence and quality strings
                    self.sequence = self.sequence.replace("\n", "")
                    self.quality = self.quality.replace("\n", "")
                    self.identifier = self.identifier
                    entry = self.to_dict()
                    entry = AttrDict(**entry)
                    try:
                        self.entries.append(entry)
                    except:
                        # not very good design but if it fails, we assume that
                        # this is the SingleFASTQ class otherwise, the FASTQ class
                        break
                    self._init()



    def _get_quality(self):
        return self._quality
    def _set_quality(self, val):
        # could be a string of characters
        # or list of integers that must be converted to characters
        if val is None:
            self._quality = None
        elif isinstance(val, str):
            self._quality = val
        elif isinstance(val, list) and len(val):
            if isinstance(val[0], int):
                self._quality = self.integer_to_quality(val)
            else:
                raise ValueError("expects a list of integers")
        else:
            raise ValueError("expects a list of integers or a string")
    quality = property(_get_quality, _set_quality)

    def check(self):
        if len(self.quality) != len(self.sequence):
            msg = "Length of the sequence and quality must be equal"
            msg += "Found length of {0} and {1}".format(len(self.sequence), len(self.quality))
            raise ValueError(msg)

    #slicing current entry
    def __getitem__(self, val):
        """you can slice the sequence f = FASTA(data)
        newf = f[5:15]

        """
        fastq = self.copy()
        fastq.sequence = fastq.sequence[val]
        fastq.quality = fastq.quality[val]
        return fastq

    def __len__(self):
        if self.sequence:
            return len(self.sequence)
        else:
            return 0

    def to_dict(self):
        """Transform current entry into a dict"""
        self.check()
        d = {}
        d['sequence'] = self.sequence[:]
        d['identifier'] = self.identifier[:]
        d['quality'] = self.quality[:]
        return d

    def _init(self):
        self.sequence = ''
        self.quality = ''
        # initialise the parsing
        self._parsing_mode = 'identifier'
        self.nolines = 0

    @classmethod
    def from_dict(cls, data):
        """Construct SingleFASTQ from dict of array-like or dicts

        :param dict data:
            {field : array-like} or {field : dict}
        """
        fastq = cls()
        fastq.sequence = data['sequence'][:]
        fastq.identifier = data['identifier'][:]
        fastq.quality = data['quality'][:]
        fastq.check()
        return fastq

    def __str__(self):
        txt = 'Identifier: ' 
        if self.identifier is not None:
            txt += self.identifier
        txt += '\nSequence:'
        if self.sequence is not None:
            txt += self.sequence
        txt += "\nQuality:" 
        if self.quality is not None:
            txt += self.quality
        return txt




class FASTQ(SingleFASTQ):
    """class to manipulate FASTQ format 
    
    WIKIPEDIA:  FASTQ is a text-based format for storing both a biological sequence (usually nucleotide sequence) and its corresponding quality scores. Both the sequence letter and quality score are encoded with a single ASCII character for brevity. It was originally developed at the Wellcome Trust Sanger Institute to bundle a FASTA sequence and its quality data, but has recently become the de facto standard for storing the output of high throughput sequencing instruments such as the Illumina Genome Analyzer.[1]

    A FASTQ file normally uses four lines per sequence.

    #. Line 1 begins with a '@' character and is followed by a sequence identifier and an optional description (like a FASTA title line).
    #. Line 2 is the raw sequence letters.
    #. Line 3 begins with a '+' character and is optionally followed by the same sequence identifier (and any description) again.
    #. Line 4 encodes the quality values for the sequence in Line 2, and must contain the same number of symbols as letters in the sequence.

    A FASTQ file containing a single sequence might look like this::

        @SEQ_ID
        GATTTGGGGTTCAAAGCAGTATCGATCAAATAGTAAATCCATTTGTTCAACTCACAGTTT
        +
        !''*((((***+))%%%++)(%%%%).1***-+*''))**55CCF>>>>>>CCCCCCC65

    The character '!' represents the lowest quality while '~' is the highest. 
    Here are the quality value characters in left-to-right increasing order of quality (ASCII):

    ::
    
        !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~

    The original Sanger FASTQ files also allowed the sequence and quality strings 
    to be wrapped (split over multiple lines), but this is generally discouraged 
    as it can make parsing complicated due to the unfortunate choice of "@" and 
    "+" as markers (these characters can also occur in the quality string).

    Here is a simple example::

        @test
        TTTCTTC
        +
        IIIIII1

    But a more ambiguous one would wrap sequence and quality and quality string may start with
    the @ or + sign. The line starting with + may be followed by comments...
    ::
    
        @test_wrap
        GAGACCTTC
        CCTAAATAC
        +test2
        @IIIIIIII
        +CII+III;
        @test_single
        TTTT
        +
        ;;;;


    f = FASTQ('filename.fastq')
    f.identifier, f.sequence, f.quality contains the first entry 

    If several entries in the fastq files, entries are provided, there are stored
    as dictionaries inside a list of entries::

        f.entries[0]['identifier']
        f.entries[0].identifier

    By default, aliases are set for the first entry:

        >>> f.identifier == f.entries[0]['identifier']
        True

    But the default entry can be set to any entries to be found in the list of entries::

        >>> f.set_index(-1) # for the last one 

    Some methods takes all entries (e.g., quality_boxplot), some others use
    only the content of the 3 attributes f.identifier, f.sequence, f.quality (e.g., the print method)



    .. references:: 
        - resources: http://en.wikipedia.org/wiki/FASTQ_format
        - P.J.A.Cock (Biopython), C.J.Fields (BioPerl), N.Goto (BioRuby),
          M.L.Heuer (BioJava) and P.M. Rice (EMBOSS).
          Nucleic Acids Research 2010 38(6):1767-1771
          http://dx.doi.org/10.1093/nar/gkp1137
    """
    # ! is the smallest quality
    # ~ is the highest quality
    # The Sanger style FASTQ files uses PHRED scores and an ASCII offset of 33 
    # (same in NCBI Short Read Archive and Illumina 1.8+
    # PHRED scores goes from 0 to 93.

    _multiple_fastq_example = """@EAS54_6_R1_2_1_413_324
CCCTTCTTGTCTTCAGCGTTTCTCC
+
;;3;;;;;;;;;;;;7;;;;;;;88
@EAS54_6_R1_2_1_540_792
TTGGCAGGCCAAGGCCGATGGATCA
+
;;;;;;;;;;;7;;;;;-;;;3;83
@EAS54_6_R1_2_1_443_348
GTTGCTTCTGGCGTGGGTGGGGGGG
+
;;;;;;;;;;;9;7;;.7;393333"""

    def __init__(self, data=None):
        self.entries = [] # used in parent class in _parse_data
        print("API in progress. May change soon")
        super(FASTQ, self).__init__(data=data)
        # if input is a dict, we should also populate the entries
        #self.to_dict()

    def _parse_data(self, data):
        super(self)._parse_data()
        self.set_entry()
    
    def set_entry(self, index=0):
        """Set the current entry given the index

        """
        self.identifier = self.entries[index]['identifier'][:]
        self.sequence = self.entries[index]['sequence'][:]
        self.quality = self.entries[index]['quality'][:]

    def __add__(self, other):
        result = self.copy()
        result.entries += [x.copy() for x in other.entries]
        return result

    def remove_duplicates(self):
        raise NotImplementedError

    @staticmethod
    def quality_from_error_probability(pe):
        """

        A quality value Q is an integer mapping of :math:`pe` (i.e., the probability 
        that the corresponding base call is incorrect). Two different equations have 
        been in use but we will only use the standard Sanger variant 
        otherwise known as Phred quality score:

        The PHRED software assigns a non-negative quality value 
        to each base using a logged transformation of the error probability:
        
        :math:`Q = -10 \log_{10}( P_e )`
        
        ::

            >>> Pe = 1e-9
            >>> Q = quality_from_error_probability(Pe)
            >>> Q
            90
            >>> quality_from_error_probability(0.1)
            0.1
            >>> quality_from_error_probability(0)
            1
            

        """
        assert pe >=0 and pe <= 1, "probability expected to be between 0 and 1"
        return -10 * pylab.log10(pe)

    def plot(self, logy=True, **kargs):
        """

        :param kargs: any argument accepted by pandas.DataFrame.plot method
        """

        df = pd.DataFrame({
            'Sequence': list(self.sequence), 
            'Quality':self.get_quality_integer()})
        df['Pe'] = df.Quality.apply(lambda x: self.error_probability_from_quality(x))
        df.Pe.plot(marker='o', **kargs)
        #for index in df.index:
        #    pylab.text(index, df['Pe'].ix[index]*1.1, df['Sequence'].ix[index])

        N = len(self)
        pylab.xlim([0, N])
        pylab.xticks(range(0, N), df.Sequence)
        pylab.title("Quality over the sequence (error probability)")
        pylab.ylabel("Probability Error")
        pylab.xlabel("Sequence")
        if logy:
            pylab.semilogy()

    @staticmethod
    def error_probability_from_quality(val):
        """Convert an error probability to PHRED quality (range 0 to about 90)
        
        .. math:: P_e = 10 ^ {-Q / 10}

        """
        import numpy as np
        val = np.array(val)
        return 10 ** (-val / 10.)

    def get_quality(self, p):
        return self.quality_from_error_probability(p)

    def get_quality_integer(self):
        """Return the list of quality values as a list of integers
        
        .. seealso:: :math:`quality_to_integer`
            
        """
        return self.quality_to_integer(self.quality, self.offset)

    @staticmethod
    def integer_to_quality(data, offset=33):
        return [chr(x + offset) for x in data]
        pass

    @staticmethod
    def quality_to_integer(data, offset=33):
        return [ord(x) - offset for x in data]


    def __len__(self):
        return len(self.entries)

    def to_qual(self):
        """
        
        ::

            >>> f = FASTQ()
            >>> f.identifier = 'TEST'
            >>> f.quality = '!@;A'
            >>> f.sequence = 'CCCC'
            >>> print(f.to_qual())
            >TEST
            0 31 26 32
        
        """
        txt = ">" + self.identifier + "\n"
        txt += " ".join([str(x) for x in self.get_quality_integer()])
        txt += "\n"
        return txt

    def clear(self):
        """Remove all entries and reset the current entry

        """
        self.entries = []
        self.identifier = ''
        self.sequence = ''
        self.quality = ''

    def copy(self):
        f = FASTQ()
        f.entries = copy.deepcopy(self.entries)
        f.identifier = self.identifier[:]
        f.sequence = self.sequence[:]
        f.quality = self.quality[:]
        #f.set_entry(0)
        try:
            f.data = self.data[:]
        except:
            pass
        return f


    def to_json(self):
        import json
        raise NotImplementedError

    def to_fasta(self):
        raise NotImplementedError

    def remove_low_quality_entries(self, mean_quality):
        df = self.get_quality_frame()
        df = df[df.mean() < mean_quality]
        indices = df.index
        self.entries = [e for i,e in enumerate(self.entries) if i in indices]

    def mask_low_quality(self, threshold):
        raise NotImplementedError

    def quality_hist(self):
        raise NotImplementedError

    def quality_to_pe(self, quality):
       pe = self.error_probability_from_quality(self.quality_to_integer(quality))
       return pe

    def get_quality_frame(self):
        qualities = [self.quality_to_integer(e.quality) for e in self.entries]
        df = pd.DataFrame({i : self.quality_to_pe(e.quality) 
            for i,e in enumerate(self.entries)})
        return df

    def quality_boxplot(self):
        """Boxplot of the quality values for each nucleotide

        one boxplot per sequence
        """
        #assume sequences have same length
        df = self.get_quality_frame() 
        df.boxplot(rot=90)
        pylab.semilogy()
        return df

    def hist_length(self, **kargs):
        df = pd.DataFrame({'Length': [len(e['sequence']) for e in self.entries]})
        df.hist(**kargs)

    def create_random_data(self, N, min_length=10, max_length=40):
        import numpy as np    
        letters = 'ACGT'
        self.clear()
        for i in range(0, N):
            d = {}
            d['identifier'] = str(i)
            mid = (max_length + min_length) / 2. 
            nseq = int(np.random.normal(mid, (max_length-min_length)/10.))
            
            d['sequence'] = ''.join([letters[np.random.randint(4)] for x in range(0, nseq)])
            d['quality'] = ''.join([self._quality_character[np.random.randint(90)] for x in range(0, nseq)])
            self.entries.append(AttrDict(**d))
            

    def add_entry(self):
        #could be entry, dict, SingleFASTQ
        raise NotImplementedError



"""DNA sequence"""
import string

"""
import os
import ctypes
#            gcc -O2 complement.c  --shared -Wl,-soname,complement 
#                -rdynamic -o complement.so -fPIC -I/usr/include/python2.7 
from os.path import join as pj

p = os.path.abspath(os.path.dirname(__file__))
try:
    cseqkit = ctypes.CDLL("complement.so")
    cseqkit.dna_complement.restype = ctypes.c_char_p
except:
    try:
        cseqkit = ctypes.cdll.LoadLibrary(pj(p, "complement.so"))
        cseqkit.dna_complement.restype = ctypes.c_char_p
    except:
        print("coudl not install lib")
"""
from biokit.sequence.seq import Sequence


__all__ = ['DNA']


class DNA(Sequence):
    """a DNA :class:`~biokit.sequence.seq.Sequence`. 

    You can add DNA sequences together::

        >>> from biokit import DNA
        >>> s1 = DNA('ACGT')
        >>> s2 = DNA('AAAA')
        >>> s1 + s2
        Sequence: ACGTAAAA (length 8)


    """
    def __init__(self, data=''):
        super(DNA, self).__init__(data)
        self.symbols = 'ACGTacgt'
        try:
            self._translate = string.maketrans('ACGTacgt', 'TGCAtgca')
        except:
            self._translate = bytes.maketrans(b'ACGTacgt', b'TGCAtgca')
        self._type = 'DNA'

    def get_complement(self):
        compl = self._data.translate(self._translate)
        return DNA(compl)
    complement = property(get_complement)

    def get_reverse_complement(self):
        complement = self.get_complement()
        return DNA(complement._data[::-1])
    reverse_complement = property(get_reverse_complement)
    
    #def _get_complement_in_c(self):
    #    return cseqkit.dna_complement(self.sequence, len(self.sequence))
    #complement2 = property(_get_complement_in_c)
    #
    #def get_complement_c(self):
    #    print("Experimetal. Not faster than Python...")
    #    return cseqkit.dna_complement(self._data, self._N)

    def gc_content(self, letters='CGS'):
        """Returns the G+C content in percentage.

        Copes mixed case sequences, and with the ambiguous nucleotide S (G or C)
        when counting the G and C content.  

        ::

            >>> from biokit.sequence.dna import DNA
            >>> d = DNA("ACGTSAAA")
            >>> d.gc_content()
            0.375

        
        """
        if len(self) == 0:
            denom = 1.
        else:
            denom = float(self._N)
        letters = [x.upper() for x in letters] + [x.lower() for x in letters]
        letters = list(set(letters))
        counter = sum(self._data.count(x) for x in letters)
        return 100. * counter / denom

    def get_rna(self):
        from biokit.sequence.rna import RNA
        # here a copy is made
        seq = self._data.replace('T', 'U')
        seq = seq.replace('t', 'u')
        return RNA(seq)








"""

Hist2d using Pandas dataframe as input
==========================================


"""



#################################################"
# wrapping of pylab hist2d function.
import pylab

from biokit import viz

X = pylab.randn(10000)
Y = pylab.randn(10000)


import pandas as pd
df = pd.DataFrame({'X':X, 'Y':Y})


from biokit.viz import hist2d
h = hist2d.Hist2D(df)
res = h.plot(bins=[40,40], contour=True, nnorm='log', Nlevels=6)





import pandas as pd
import pylab



class Boxplot(object):

    def __init__(self, data):

        # if data is a dataframe, keep it else, transform to dataframe
        try:
            self.df = pd.DataFrame(data)
        except:
            self.df = data

        self.xmax = self.df.shape[1]
        self.X =  None

    def plot(self, color_line='r', bgcolor='grey', color='yellow', lw=4, 
            hold=False, ax=None):

        xmax = self.xmax + 1
        if ax:
            pylab.sca(ax)
        pylab.fill_between([0,xmax], [0,0], [20,20], color='red', alpha=0.3)
        pylab.fill_between([0,xmax], [20,20], [30,30], color='orange', alpha=0.3)
        pylab.fill_between([0,xmax], [30,30], [41,41], color='green', alpha=0.3)

        if self.X is None:
            X = range(1, self.xmax + 1)

        pylab.fill_between(X, 
            self.df.mean()+self.df.std(), 
            self.df.mean()-self.df.std(), 
            color=color, interpolate=False)

        pylab.plot(X, self.df.mean(), color=color_line, lw=lw)
        pylab.ylim([0, 41])
        pylab.xlim([0, self.xmax+1])
        pylab.title("Quality scores across all bases")
        pylab.xlabel("Position in read (bp)")
        pylab.ylabel("Quality")
        pylab.grid(axis='x')



"""Allows mouse wheel to work within matplotlib


http://stackoverflow.com/questions/11551049/matplotlib-plot-zooming-with-scroll-wheel
"""
from matplotlib.pyplot import figure, show
import numpy

class ZoomPan:
    def __init__(self):
        self.press = None
        self.cur_xlim = None
        self.cur_ylim = None
        self.x0 = None
        self.y0 = None
        self.x1 = None
        self.y1 = None
        self.xpress = None
        self.ypress = None

    def zoom_factory(self, ax, base_scale = 2.):
        def zoom(event):
            cur_xlim = ax.get_xlim()
            cur_ylim = ax.get_ylim()

            xdata = event.xdata # get event x location
            ydata = event.ydata # get event y location

            if event.button == 'down':
                # deal with zoom in
                scale_factor = 1 / base_scale
            elif event.button == 'up':
                # deal with zoom out
                scale_factor = base_scale
            else:
                # deal with something that should never happen
                scale_factor = 1
                print(event.button)

            new_width = (cur_xlim[1] - cur_xlim[0]) * scale_factor
            new_height = (cur_ylim[1] - cur_ylim[0]) * scale_factor

            relx = (cur_xlim[1] - xdata)/(cur_xlim[1] - cur_xlim[0])
            rely = (cur_ylim[1] - ydata)/(cur_ylim[1] - cur_ylim[0])

            ax.set_xlim([xdata - new_width * (1-relx), xdata + new_width * (relx)])
            ax.set_ylim([ydata - new_height * (1-rely), ydata + new_height * (rely)])
            ax.figure.canvas.draw()

        fig = ax.get_figure() # get the figure of interest
        fig.canvas.mpl_connect('scroll_event', zoom)

        return zoom

    def pan_factory(self, ax):
        def onPress(event):
            if event.inaxes != ax: return
            self.cur_xlim = ax.get_xlim()
            self.cur_ylim = ax.get_ylim()
            self.press = self.x0, self.y0, event.xdata, event.ydata
            self.x0, self.y0, self.xpress, self.ypress = self.press

        def onRelease(event):
            self.press = None
            ax.figure.canvas.draw()

        def onMotion(event):
            if self.press is None: return
            if event.inaxes != ax: return
            dx = event.xdata - self.xpress
            dy = event.ydata - self.ypress
            self.cur_xlim -= dx
            self.cur_ylim -= dy
            ax.set_xlim(self.cur_xlim)
            ax.set_ylim(self.cur_ylim)

            ax.figure.canvas.draw()

        fig = ax.get_figure() # get the figure of interest

        # attach the call back
        fig.canvas.mpl_connect('button_press_event',onPress)
        fig.canvas.mpl_connect('button_release_event',onRelease)
        fig.canvas.mpl_connect('motion_notify_event',onMotion)

        #return the function
        return onMotion


def example():
    fig = figure()

    ax = fig.add_subplot(111, xlim=(0,1), ylim=(0,1), autoscale_on=False)

    ax.set_title('Click to zoom')
    x,y,s,c = numpy.random.rand(4,200)
    s *= 200
    
    ax.scatter(x,y,s,c)
    scale = 1.1
    zp = ZoomPan()
    figZoom = zp.zoom_factory(ax, base_scale = scale)
    figPan = zp.pan_factory(ax)
    show()




"""RNA sequence"""
import string
from biokit.sequence.seq import Sequence


__all__ = ['RNA']



class RNA(Sequence):
    def __init__(self, sequence=''):
        super(RNA, self).__init__(sequence)
        self.symbols = 'ACGUacgu'
        try:
            self._translate = string.maketrans('ACGUacgu', 'UGCAugca')
        except:
            self._translate = bytes.maketrans(b'ACGUacgu', b'UGCAugca')
        self._type = 'RNA'

    def get_complement(self):
        compl = self._data.translate(self._translate)
        return RNA(compl)
    complement = property(get_complement)

    def get_reverse_complement(self):
        complement = self.get_complement()
        return RNA(complement._data[::-1])
    reverse_complement = property(get_reverse_complement)

    def get_dna(self):
        # here a copy is made
        from biokit.sequence.dna import DNA
        seq = self._data.replace('U', 'T')
        seq = seq.replace('u', 't')
        return DNA(seq)

    def gc_content(self, letters='CGS'):
        """Returns the G+C content in percentage.

        Copes mixed case sequences, and with the ambiguous nucleotide S (G or C)
        when counting the G and C content.

        ::

            >>> from biokit.sequence.dna import RNA
            >>> d = RNA("ACGTSAAA")
            >>> d.gc_content()
            0.375


        """
        if len(self) == 0:
            denom = 1.
        else:
            denom = float(self._N)
        letters = [x.upper() for x in letters] + [x.lower() for x in letters]
        letters = list(set(letters))
        counter = sum(self._data.count(x) for x in letters)
        return 100. * counter / denom




# import collections
import string

from biokit.sequence.seq import Sequence


class Peptide(Sequence):
    """a Peptide :class:`~biokit.sequence.seq.Sequence`. 

    You can add Peptide sequences together::

        >>> from biokit import DNA
        >>> s1 = Peptide('ACGT')
        >>> s2 = Peptide('AAAA')
        >>> s1 + s2
        Sequence: ACGTAAAA (length 8)

    .. note:: redundant with Sequence but may evolve in the future.
    """
    def __init__(self, data):
        super(Peptide, self).__init__(data)
        self.symbols = 'ACGTacgt'
        self._translate = string.maketrans('ACGTacgt', 'TGCAtgca')







#!/usr/bin/env python
"""FORK of PyperR version 1.1.2 from pypi

Reason for the fork. Code are written on top of pyper inside biokit.
Yet, PypeR is failing sometimes (see examples below) and we may need to
hack it. It had also Python 3 compat issues.

Examples that fail:

    installed.packages()  --> numpy.array failing
    packageVersion("CellNOptR")  --> special unicode not interpreted

The get() function fails with S4 objects such as CNOlist
We return None to avoid the failure::

    else if (is.object(x)) 'None'

----------------------------------------------------------------------------

         PypeR (PYthon-piPE-R)

PypeR is free software subjected to the GPL license 3.0. and comes with
ABSOLUTELY NO WARRANT. This package provides a light-weight interface to use R
in Python by pipe.  It can be used on multiple platforms since it is written in
pure python.


Usage:
    The usage of this packages is very simple. Examples are presented in the
    file "test.py" in the distribution package.

    PypeR provide a class "R" to wrap the R language. An instance of the R
    class is used to manage an R process. Different instances can use different
    R installations. On POSIX systems (including the Cygwin environment on
    Windows), it is even possible to use an R installed on a remote computer.

    Basicly, there are four ways to use an instance of the R class.

    1. Use the methods of the instance
        methods include:
            run:This method is used to pass an R command string to the R process,
                the return value is a string - the standard output from R. Note
                that the return value usually includes the R expression (a
                series of R codes) themselves and the output of the R
                expression.  If the real result value is wanted, use the
                function "get" instead.
            assign: Assign a value to an R variable. No return value.
            get: Get the result of an R expression.
            remove: Remove a R variable.

    2. Call the instance as a function
        The instance is callable. If called as a function, it behaves just
        same as its "run" method.

    3. Use the instance as a Python dictionary
        The instance can mimic some operations on a python dictionary,
        typically, to assign values to R variables, to retrieve values for any
        R expression, or delete an R variable. These two operations do same
        jobs as the methods "assign", "get", and "remove".

    4. Access R variables as if they are the attributes of the instance.
        If the variable name cannot be found in the instance or its class, the
        instance will try to get/set/remove it in R. This way is similar to 3,
        but with more limitations, e.g., the R variable name cannot contain any
        DOT (.)

    Considering that any code block in R is an expression, the "get" method (or
    the form of retrieving values from a dictionary) can be used to run a
    number of R commands with the final result returned.

    Note that PypeR do NOT validate/convert a variable name when pass it to R.
    If a variable name with a leading underscore ("_"), although it legal in
    python, is passed to R, an RError will be raised.

Conversions:
    Python -> R
        None -> NULL, NaN -> NaN, Inf -> Inf
    R -> Python (numpy)
        NULL -> None, NA -> None, NaN -> None (NaN), Inf -> None (Inf)

DEBUG model:
    Since the child process (R) can be easily killed by any ocassional error in
    the codes passed to it, PypeR is set to "DEBUG" model by default. This
    means that any code blocks send to R will be wrapped in the function
    "try()", which will prevent R from crashing. To disable the "DEBUG" model,
    the user can simple set the variable "DEBUG_MODE" in the R class or in its
    instance to False.

    To model the behavior of the "get" method of a Python dictionary, the
    method "get" allows wild values for variables that does not exists in R.
    Then the R expression will always be wrapped in "try()" to avoid R crashing
    if the method "get" is called.
"""

import os
import sys
import time
import re
import tempfile
from types import *   # need to clean that
import subprocess

from biokit.rtools.r4python import r4python
from biokit.rtools.py2r import *

# Even though there are not used in this module, these two imports
# are required for the RSession to properly interpret the strings
# as objects
import numpy
import pandas

if sys.version < '3.0':
    _mystr = _mybytes = lambda s: s
    _in_py3 = False
else:
    from functools import reduce
    long, basestring, unicode = int, str, str
    _mybytes = lambda s: bytes(s, 'utf8')  # 'ascii')
    _mystr = lambda s: str(s, 'utf8')
    _in_py3 = True


_has_subp = False



PIPE, _STDOUT = None, None
def Popen(CMD, *a, **b):
    #original pyper code:
    #class A:
    #    None
    #p = A()
    #p.stdin, p.stdout = os.popen4(' '.join(CMD))
    # popen4 does not work in Python 3, so we use a Popen
    p = subprocess.Popen(CMD,shell=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            close_fds=True)
    return(p)

def sendAll(p, s):
    p.stdin.write(_mybytes(s))
    #os.write(p.stdin.fileno(), s)
    p.stdin.flush()

def readLine(p, dump_stdout=False, *a, **b):
    #rv = _mystr(nbsr.readline())
    rv = _mystr(p.stdout.readline())
    if dump_stdout:
        sys.stdout.write(rv)
        sys.stdout.flush()
    return(rv)


class RError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return(repr(self.value))


class R(object):
    # "del r.XXX" fails on FePy-r7 (IronPython 1.1 on .NET 2.0.50727.42) if using old-style class
    """A Python class to enclose an R process."""

    __Rfun = r4python
    _DEBUG_MODE = True

    def __init__(self, RCMD='R', max_len=10000, use_dict=None,
                 host='localhost', user=None, ssh='ssh', return_err=True, dump_stdout=False,
                 verbose=False, options=('--quiet', '--no-save', '--no-restore')):
        '''
        RCMD: The name of a R interpreter, path information should be included
            if it is not in the system search path.
        use_dict: A R named list will be returned as a Python dictionary if
            "use_dict" is True, or a list of tuples (name, value) if "use_dict"
            is False. If "use_dict" is None, the return value will be a
            dictionary if there is no replicated names, or a list if replicated
            names found.
        host: The computer name (or IP) on which the R interpreter is
            installed. The value "localhost" means that R locates on the the
            localhost computer. On POSIX systems (including Cygwin environment
            on Windows), it is possible to use R on a remote computer if the
            command "ssh" works. To do that, the user needs to set this value,
            and perhaps the parameter "user".
        user: The user name on the remote computer. This value needs to be set
            only if the user name on the remote computer is different from the
            local user. In interactive environment, the password can be input
            by the user if prompted. If running in a program, the user needs to
            be able to login without typing password!
        ssh: The program to login to remote computer.
        return_err: redirect stderr to stdout
        dump_stdout:
            prints output from R directly to sys.stdout, useful for long running
            routines which print progress during execution.
        '''
        # use self.__dict__.update to register variables since __setattr__ is
        # used to set variables for R.  tried to define __setattr in the class,
        # and change it to __setattr__ for instances at the end of __init__,
        # but it seems failed.
        # -- maybe this only failed in Python2.5? as warned at
        # http://wiki.python.org/moin/NewClassVsClassicClass:
        # "Warning: In 2.5, magic names (typically those with a double
        # underscore (DunderAlias) at both ends of the name) may look at the
        # class rather than the instance even for old-style classes."
        self.__dict__.update({
            'prog': None,
            'Rfun': self.__class__.__Rfun,
            'Rexecutable': RCMD,
            'max_len': max_len,
            'use_dict': use_dict,
            'verbose': verbose,
            'dump_stdout': dump_stdout,
            'localhost': host == 'localhost',
            'newline': sys.platform == 'win32' and '\r\n' or '\n',
            'sendAll' : sendAll # keep a reference to the global function "sendAll" which will be used by __del__
            })

        RCMD = [RCMD]
        if not self.localhost:
            RCMD.insert(0, host)
            if user:
                RCMD.insert(0, '-l%s' % user)
            RCMD.insert(0, ssh)

        for arg in options:
            if arg not in RCMD:
                RCMD.append(arg)

        if _has_subp and hasattr(subprocess, 'STARTUPINFO'):
            info = subprocess.STARTUPINFO()
            try:
                if hasattr(subprocess, '_subprocess'):
                    info.dwFlags |= subprocess._subprocess.STARTF_USESHOWWINDOW
                    info.wShowWindow = subprocess._subprocess.SW_HIDE
                else:
                    info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                    info.wShowWindow = subprocess.SW_HIDE
            except:
                info = None
        else:
            info = None

        # create stderr to replace None for py2exe:
        # http://www.py2exe.org/index.cgi/Py2ExeSubprocessInteractions
        if sys.platform != 'win32':
            childstderr = None
        else:
            if hasattr(sys.stderr, 'fileno'):
                childstderr = sys.stderr
            elif hasattr(sys.stderr, '_file') and hasattr(sys.stderr._file, 'fileno'):
                childstderr = sys.stderr._file
            else:  # Give up and point child stderr at nul
                childstderr = file('nul', 'a')

        from easydev import AttrDict
        self.__dict__['subprocess_args'] = AttrDict(**{
                'RCMD': RCMD,
                'PIPE': PIPE,
                'stderr': return_err and _STDOUT or childstderr,
                'info': info})

        self.reconnect()

    def reconnect(self):
        """TC: Nov 2014

        If CTRL+C is called, the pipe is broken, in which case, reconnecting
        should be called.

        """
        args = self.subprocess_args
        RCMD = args['RCMD']
        PIPE = args['PIPE']
        stderr = args['stderr']
        info = args['info']
        self.__dict__['prog'] = Popen(RCMD, stdin=PIPE, stdout=PIPE,
                stderr=stderr, startupinfo=info)
        self.__call__(self.Rfun)

    def __runOnce(self, CMD, use_try=None):
        """CMD: a R command string"""


        use_try = use_try or self._DEBUG_MODE
        newline = self.newline
        tail_token = 'R command at time: %s' % repr(time.time())
        # tail_token_r = re.sub(r'[\(\)\.]', r'\\\1', tail_token)
        tail_cmd = 'print("%s")%s' % (tail_token, newline)
        tail_token = tail_token.replace(' ', '\\s').replace('.', '\\.').replace('+', '\\+')
        re_tail = re.compile(r'>\sprint\("%s"\)\r?\n\[1\]\s"%s"\r?\n$' % (tail_token, tail_token))

        if len(CMD) <= self.max_len or not self.localhost:
            fn = None
            CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD.replace('\\', '\\\\'),
                    newline, tail_cmd)
        else:
            fh, fn = tempfile.mkstemp()
            os.fdopen(fh, 'wb').write(_mybytes(CMD))
            if sys.platform == 'cli':
                os.close(fh)  # this is necessary on IronPython
            fn = fn.replace('\\', '/')

            params = {'fn':fn , 'newline':newline, 'tail_cmd':tail_cmd}
            if use_try is True:
                CMD = 'try({source("%(fn)s")})%(newline)s ' % params
                CMD += 'dummy=file.remove(%(fn)r)%(newline)s%(tail_cmd)s' % params
            else:
                CMD = '({source("%(fn)s")})%(newline)s ' % params
                CMD += 'dummy=file.remove(%(fn)r)%(newline)s%(tail_cmd)s' % params

        try:
            self.sendAll(self.prog, CMD)
        except IOError:
            if self.verbose:
                print("PIPE was broken. Reconnecting...")
            self.reconnect()
            self.sendAll(self.prog, CMD)
        except Exception as err:
            print("Failed")
            raise err


        rlt = ''
        while not re_tail.search(rlt):
            try:
                rltonce = readLine(self.prog, dump_stdout=self.dump_stdout)
                if rltonce:
                    rlt = rlt + rltonce
            except:
                break
        else:
            rlt = re_tail.sub('', rlt)
            if rlt.startswith('> '):
                rlt = rlt[2:]
        return(rlt)

    def __call__(self, CMDS=[], use_try=None):
        """Run a (list of) R command(s),

        :param list CMDS: either a list of commands or a single command as a string.
        :return: nothing but debugLevel is filled

        """
        if isinstance(CMDS, basestring):  # a single command
            rlt = self.__runOnce(CMDS, use_try=use_try)
        else:
            rlt = self.__runOnce('; '.join(CMDS), use_try=use_try)
        return rlt

    def __getitem__(self, obj, use_try=None, use_dict=None): # to model a dict: "r['XXX']"
        """Get the value of an R variable or expression.


        :return: a Python object.

        :param obj: a string - the name of an R variable, or an R expression
        :param use_try: use "try" function to wrap the R expression. This can avoid R
            crashing if the obj does not exist in R.
        :param use_dict: named list will be returned a dict if use_dict is True,
            otherwise it will be a list of tuples (name, value)
        """
        if obj.startswith('_'):
            raise RError('Leading underscore ("_") is not permitted in R variable names!')
        use_try = use_try or self._DEBUG_MODE
        if use_dict is None:
            use_dict = self.use_dict
        cmd = '.getRvalue4Python__(%s, use_dict=%s)' % (obj, use_dict is None and 'NULL' or use_dict and 'TRUE' or 'FALSE')
        rlt = self.__call__(cmd, use_try=use_try)
        head = (use_try and 'try({%s})%s[1] ' or '%s%s[1] ') % (cmd, self.newline)
        # sometimes (e.g. after "library(fastICA)") the R on Windows uses '\n' instead of '\r\n'
        head = rlt.startswith(head) and len(head) or len(head) - 1
        tail = rlt.endswith(self.newline) and len(rlt) - len(self.newline) or len(rlt) - len(self.newline) + 1 # - len('"')
        try:
            rlt = eval(eval(rlt[head:tail])) # The inner eval remove quotes and recover escaped characters.
        except:
            try:
                # some hack for installed.packages()
                code = rlt[head:tail]
                code = code.replace('\\\\"', '"')
                #code = code.replace("\n", "")
                #print(code)
                rlt = eval(code)  # replace \\" with "
            except Exception as e:
                print(e)

                raise RError(rlt)
        return(rlt)

    def __setitem__(self, obj, val):  # to model a dict: "r['XXX'] = YYY"
        """ Assign a value (val) to an R variable (obj).

        :param obj: a string - the name of an R variable
        :param val: a python object - the value to be passed to an R object
        """
        if obj.startswith('_'):
            raise RError('Leading underscore ("_") is not permitted in R variable names!')
        self.__call__('%s <- %s' % (obj, Str4R(val)))

    def __delitem__(self, obj):  # to model a dict: "del r['XXX']"
        if obj.startswith('_'):
            raise RError('Leading underscore ("_") is not permitted in R variable names!')
        self.__call__('rm(%s)' % obj)

    def __del__(self):  # to model "del r"
        # if we do not have those 2 lines, prog may not be in the dictionary
        # and we enter in an infite recursion loop ....
        if 'prog' not in self.__dict__.keys():
            return

        if self.prog:
            try:
                self.sendAll(self.prog, 'q("no")'+self.newline)
            except:
                pass
            self.prog = None

    def __getattr__(self, obj, use_dict=None):  # to model object attribute: "r.XXX"
        """
        :param obj: a string - the name of an R variable
        :param use_dict: named list will be returned a dict if use_dict is True,
            otherwise it will be a list of tuples (name, value)
        """
        # Overriding __getattr__ is safer than __getattribute__ since it is
        # only called as a last resort i.e. if there are no attributes in the
        # instance that match the name
        if obj in self.__dict__:
            return(self.__dict__[obj])
        if obj in self.__class__.__dict__:
            return(self.__class__.__dict__[obj])
        try:
            if use_dict is None:
                use_dict = self.use_dict
            rlt = self.__getitem__(obj, use_dict=use_dict)
        except:
            raise
        return(rlt)

    def __setattr__(self, obj, val):  # to model object attribute: "r.XXX = YYY"
        if obj in self.__dict__ or obj in self.__class__.__dict__: # or obj.startswith('_'):
            self.__dict__[obj] = val # for old-style class
            #object.__setattr__(self, obj, val) # for new-style class
        else:
            self.__setitem__(obj, val)

    def __delattr__(self, obj):  # to model object attribute: "del r.XXX"
        if obj in self.__dict__:
            del self.__dict__[obj]
        else:
            self.__delitem__(obj)

    def get(self, obj, default=None, use_dict=None):  # to model a dict: "r.get('XXX', 'YYY')"
        """
        :param obj: a string - the name of an R variable, or an R expression
        :param default: a python object - the value to be returned if failed to get data from R
        :param use_dict: named list will be returned a dict if use_dict is True,
            otherwise it will be a list of tuples (name, value). If use_dict is
            None, the value of self.use_dict will be used instead.
        """
        try:
            rlt = self.__getitem__(obj, use_try=True, use_dict=use_dict)
        except:
            if True:  # val is not None:
                rlt = default
            else:
                raise RError('No this object!')
        return(rlt)

    run, assign, remove = __call__, __setitem__, __delitem__





# -*- python -*-
#
#  This file is part of biokit software
#
#  Copyright (c) 2014-
#
#  File author(s): Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/biokit
#
##############################################################################
from biokit.rtools import pyper

__all__ = ['RSession']


class RSession(pyper.R):
    """Interface to a R session

    This class uses the pyper package to provide an access to R (via 
    a subprocess). You can call R script and get back the results
    into the session as Python objects. Returned objects may be transformed
    into numpy arrays or Pandas datafranes.

    Here is a very simple example but any complex R scripts can be provided
    inside the :meth:`run` method::

        from biokit.rtools import RSession
        session = RSession()
        session.run("mylist = c(1,2,3)")
        a = session("mylist") # access to the R object
        a.sum() # a is numpy array

    There are different ways to access to the R object::

        # getter
        session['a']
        # method-wise:
        session.get('a')
        # attribute:
        session.a

    For now, this is just to inherit from **pyper.R** class but there is no
    additional features. This is to create a common API.

    """
    def __init__(self, RExecutable='R', max_len=1000,
            use_dict=None, host='localhost', user=None, ssh='ssh',
            verbose=False, return_err=True):
        """

        :param str RCMD: the name of the R executable
        :param max_len: define the upper limitation for the length of command string. A
           command string will be passed to R by a temporary file if it is longer
           than this value.
        :param host: The computer name (or IP) on which the R interpreter is
           installed. The value "localhost" means that the R locates on the
           the localhost computer. On POSIX systems (including Cygwin
           environment on Windows), it is possible to use R on a remote
           computer if the command "ssh" works. To do that, the user need set
           this value, and perhaps the parameter "user".
        :param user: The user name on the remote computer. This value need to be set
           only if the user name is different on the remote computer. In
           interactive environment, the password can be input by the user if
           prompted. If running in a program, the user need to be able to
           login without typing password! (i.e., you need to set your SSH keys)
        :param ssh: The program to login to remote computer.
        :param kargs: must be empty. Error raised otherwise
        """
        from easydev import cmd_exists
        if host == 'localhost' and cmd_exists(RExecutable) is False:
            raise Exception('Could not find the R executable %s in your path' % RExecutable)

        super(RSession, self).__init__(RCMD=RExecutable, max_len=max_len,
            use_dict=use_dict, host=host, user=user, ssh=ssh, return_err=return_err,
            dump_stdout=verbose)


    def get_version(self):
        """Return the R version"""
        return self.version

    def __repr__(self):
        txt = "Rsession information:" 
        txt += "\nverbosity: " + str(self.verbose)
        txt += "\ndump_stdout: " + str(self.dump_stdout)
        txt += "\ndebug mode: " + str(self._DEBUG_MODE)
        return txt






import scipy.stats
import numpy as np
import pandas as pd



class ANOVA(object):
    """

    assumption: normality of the data and independence (eg. True across cell line,
        not over protein)
    """

    def __init__(self, df):

        self.df = df.copy()


    def anova(self):
        F, P = scipy.stats.f_oneway(*[self.df[x] for x in self.df.columns])
        return F, P


    def imshow_anova_pairs(self, log=True, **kargs):
        N = len(self.df.columns)

        # could use a dataframe straight way ?
        res = np.ones((N, N))
        for i,col1 in enumerate(self.df.columns):
            for j,col2 in enumerate(self.df.columns):
                d1 = self.df[col1]
                d2 = self.df[col2]
                F, P = scipy.stats.f_oneway(*[d1, d2])
                res[i][j] = P
        df = pd.DataFrame(res, index=self.df.columns, columns=self.df.columns)
        #FIXME: may have na, which are set to 1
        df = df.fillna(1)
        from biokit.viz import Imshow
        if log==True:  
            Imshow(-np.log10(df)).plot(**kargs)
        else:
            Imshow(df).plot(**kargs)
        return df




# source
# http://nbviewer.ipython.org/github/tritemio/notebooks/blob/master/Mixture_Model_Fitting.ipynb

from easydev import DevTools
devtools = DevTools()
from scipy.optimize import minimize, show_options
import scipy.stats as ss
import numpy as np
import pylab
from easydev import AttrDict

from . import criteria

import numpy as np

half_log_two_pi = 0.5*np.log(2*np.pi)


class Model(object):
    """New base model"""
    def __init__(self):
        pass
    def log_density(self, data):
        raise NotImplementedError
    def estimate(self, data, weights):
        raise NotImplementedError
    def generate(self):
        raise NotImplementedError
    def pdf(self):
        raise NotImplementedError
    def __repr__(self):
        raise NotImplementedError


class GaussianModel(Model):
    """New gaussian model not used yet"""

    def __init__(self, mu=1, sigma=1):
        super(GaussianModel, self).__init__()
        self.mu = mu
        self.sigma = sigma

    def log_density(self, data):
        # log of gaussian distribution
        res = -(data-self.mu)**2 / (2 * self.sigma**2)
        res += - np.log(self.sigma)  - half_log_two_pi
        return res

    def estimate(self, data, weights=None):
        assert(len(data.shape) == 1), "Expect 1D data!"
        # d L/ dmu ==0 and dL/dsigma==0
        if weights:
            wsum = np.sum(weights)
            self.mu = np.sum(weights * data) / wsum
            self.sigma = np.sqrt(np.sum(weights * (data - self.mu) ** 2) / wsum)
        else:
            self.mu = np.sum(data)
            self.sigma = np.sqrt(np.sum( (data - self.mu) ** 2) )

    def generate(self, N):
        return np.array([scipy.stats.norm.rvs(self.mu, self.sigma) for this in range(N)])

    def pdf(self,data):
        pass


class PoissonModel(Model):
    """New poisson model not used yet"""
    def __init__(self, lmbda=10):
        super(PoissonModel, self).__init__()
        self.lmbda = lmbda
    def log_density(self, data):
        # log of gaussian distribution
        return - self.lmbda + data * np.log(self.lmbda)
    def estimate(self, data, weights=None):
        assert(len(data.shape) == 1), "Expect 1D data!"
        # d L/ dmu ==0 and dL/dsigma==0
        if weights:
            wsum = np.sum(weights)
            self.lmbda = np.sum(weights * data) / wsum
        else:
            self.lmbda = np.sum(data)

        # Log likelihood of poisson of N iid is L = PI_i^n \lambda^{x_i}
        # e^-\lambda / x_i!  and the loglikelihood is then l = sum_i^n
        # x_i log\lmabda - n \lambda
    def generate(self, N):
        return np.array([scipy.stats.poisson.rvs(self.lmbda) for this in range(N)])
    def __repr__(self):
        return "Poisson[lmbda={lmbda:.4g}]".format(lmbda=self.lmbda)



class GaussianMixture(object):
    """Creates a mix of Gaussian distribution

    .. plot::
        :width: 80%
        :include-source:

        from biokit.stats.mixture import GaussianMixture
        m = GaussianMixture(mu=[-1,1], sigma=[0.5, 0.5], mixture=[.2, .8], N=1000)
        m.plot()

    """
    def __init__(self, mu=[-1,1], sigma=[1,1], mixture=[0.5,0.5], N=1000):
        """.. rubric:: constructor

        :param list mu: list of mean for each model
        :param list sigma: list of standard deviation for each model
        :param list mixture: list of amplitude for each model

        """
        assert len(mu) == len(sigma)
        assert len(mu) == len(mixture)
        self.mu = mu
        self.sigma = sigma
        self.mixture = mixture
        self.data = []
        self.N = N
        self.Ns = [int(x*N) for x in mixture]
        self.k = len(self.mu)

        if sum(self.Ns) != N:
            print('Warning: rounding mixture ratio. total N will be %s' %
                    sum(self.Ns))

        for m, s, n in zip(self.mu, self.sigma, self.Ns):
            data = pylab.normal(m, s, size=n)
            self.data.extend(data)

    def plot(self, bins=30, normed=True):
        self.hist(bins=bins, normed=normed)

    def hist(self, bins=30, normed=True):
        pylab.hist(self.data, bins=bins, normed=normed)



class GaussianMixtureModel(object):
    """Gaussian Mixture Model

    .. plot::

        from biokit.stats import mixture
        m = mixture.GaussianMixtureModel()
        m.pdf(1, params=[1, 0.5, 0.2, 1, 0.5, 0.8])

    """
    def __init__(self, k=2):
        self.k = k

    def pdf(self, x, params, normalise=True):
        """Expected parameters are


        params is a list of gaussian distribution ordered as mu, sigma, pi,
        mu2, sigma2, pi2, ...

        """
        assert divmod(len(params), 3)[1] == 0
        assert len(params) >= 3 * self.k
        k = len(params) / 3

        self.k = k

        pis = np.array(params[2::3])

        if any(np.array(pis)<0):
            return 0
        if normalise is True:
            pis /= pis.sum()
        # !!! sum pi must equal 1 otherwise may diverge badly
        data = 0
        for i in range(0, int(k)):
            mu, sigma, pi_ = params[i*3: (i+1)*3]
            pi_ = pis[i]
            if sigma != 0:
                data += pi_ * ss.norm.pdf(x, mu, sigma)
        return data

    def log_likelihood(self, params, sample):
        res =  -1 * pylab.log(self.pdf(sample, params)).sum()
        return res


class Fitting(object):
    """Base class for :class:`EM` and :class:`GaussianMixtureFitting`"""
    def __init__(self, data, k=2, method='Nelder-Mead'):
        """.. rubric:: constructor

        :param list data:
        :param int k: number of GMM to use
        :param str method: minimization method to be used (one of scipy optimise module)


        """
        self.data = np.array(data)
        self.size = float(len(self.data))
        self._k = k
        self._model = None
        # initialise the model
        self.k = k
        self.verbose = True

    def _get_k(self):
        return self._k
    def _set_k(self, k):
        assert k > 0
        gmm = GaussianMixtureModel(k=k)
        self._k = k
        self._model = gmm
    k = property(_get_k, _set_k)

    def _get_model(self):
        return self._model
    model = property(_get_model)

    def get_guess(self):
        """Random guess to initialise optimisation"""
        params = {}
        m = self.data.min()
        M = self.data.max()
        range_ = M - m

        mus = [m + range_ / (self.k+1.) * i for i in range(1, self.k+1)]
        params['mus'] = mus

        sigma = range_ / float(self.k+1) / 2.
        params['sigmas'] = [sigma] * self.k

        params['pis'] = [1./self.k] * self.k

        params = [ [mu,sigma,pi]  for mu,sigma,pi in
                    zip(params['mus'], params['sigmas'], params['pis'])]
        params = list(pylab.flatten(params))
        return params

    def plot(self, normed=True, N=1000, Xmin=None, Xmax=None, bins=50, color='red', lw=2,
            hist_kw={'color':'#5F9EA0', "edgecolor":"k"}, ax=None):

        if ax:
            ax.hist(self.data, normed=normed, bins=bins, **hist_kw)
        else:
            pylab.hist(self.data, density=normed, bins=bins, **hist_kw)
        if Xmin is None:
            Xmin = self.data.min()
        if Xmax is None:
            Xmax = self.data.max()
        X = pylab.linspace(Xmin, Xmax, N)

        if ax:
            ax.plot(X, [self.model.pdf(x, self.results.x) for x in X],
                    color=color, lw=lw)
        else:
            pylab.plot(X, [self.model.pdf(x, self.results.x) for x in X],
                    color=color, lw=lw)

        K = len(self.results.x)
        # The PIs must be normalised
        for i in range(self.k):

            mu, sigma, pi_ = self.results.mus[i], self.results.sigmas[i], self.results.pis[i]
            if ax:
                ax.plot(X, [pi_ * ss.norm.pdf(x, mu, sigma) for x in X], 
                        'k--', alpha=0.7, lw=2)
            else:
                pylab.plot(X, [pi_ * ss.norm.pdf(x, mu, sigma) for x in X], 
                        'k--', alpha=0.7, lw=2)


class GaussianMixtureFitting(Fitting):
    """GaussianMixtureFitting using scipy minization

    .. plot::
        :width: 80%
        :include-source:

        from biokit.stats.mixture import GaussianMixture, GaussianMixtureFitting
        m = GaussianMixture(mu=[-1,1], sigma=[0.5,0.5], mixture=[0.2,0.8])
        mf = GaussianMixtureFitting(m.data)
        mf.estimate(k=2)
        mf.plot()


    """
    def __init__(self, data, k=2, method='Nelder-Mead'):
        """

        Here we use the function minimize() from scipy.optimization.
        The list of (currently) available minimization methods is 'Nelder-Mead' (simplex),
        'Powell', 'CG', 'BFGS', 'Newton-CG',> 'Anneal', 'L-BFGS-B' (like BFGS but bounded),
        'TNC', 'COBYLA', 'SLSQPG'.

        """
        super(GaussianMixtureFitting, self).__init__(data, k=k, method=method)
        self._method = method

    def _get_method(self):
        return self._method
    def _set_method(self, method):
        devtools.check_param_in_list(method, ['Nelder-Mead',
            'Powell', 'CG', 'BFGS', 'Newton-CG', 'Anneal', 'L-BFGS-B'])
        self._method = method
    method = property(_get_method, _set_method)

    def estimate(self, guess=None, k=None, maxfev=2e4, maxiter=1e3,
            bounds=None):
        """guess is a list of parameters as expected by the model


        guess = {'mus':[1,2], 'sigmas': [0.5, 0.5], 'pis': [0.3, 0.7]  }

        """
        if k is not None:
            self.k = k

        if guess is None:
            # estimate the mu/sigma/pis from the data
            guess = self.get_guess()

        res = minimize(self.model.log_likelihood, x0=guess, args=(self.data,),
            method=self.method, options=dict(maxiter=maxiter, maxfev=maxfev),
            bounds=bounds)

        self.results = res
        pis = np.array(self.results.x[2::3])
        self.results.pis_raw = pis.copy()
        # The ratio may be negative, in which case we need to normalise.
        # An example would be to have -0.35, -0.15, which normalise would five 0.7, 0.3 as expected.
        """if sum(pis<0) > 0:
            unstable = True
            pis /= pis.sum()
            if self.verbose:
                print("Unstable... found negative pis (k=%s)" % self.k)
        else:
            unstable = False
            pis /= pis.sum()
        """
        unstable = False
        k = len(self.results.x)/3
        params = []
        for i in range(0, int(k)):
            params.append(self.results.x[i*3])
            params.append(self.results.x[(i*3+1)])
            params.append(pis[i])
        self.results.x = params

        # FIXME shall we multiply by -1 ??
        self.results.log_likelihood = self.model.log_likelihood(params, self.data)
        if self.results.log_likelihood and unstable is False:
            self.results.AIC = criteria.AIC(self.results.log_likelihood, self.k, logL=True)
            self.results.AICc = criteria.AICc(self.results.log_likelihood, self.k, self.data.size, logL=True)
            self.results.BIC = criteria.BIC(self.results.log_likelihood, self.k, self.data.size, logL=True)
        else:
            self.results.AIC = 1000
            self.results.AICc = 1000
            self.results.BIC = 1000

        pis = np.array(self.results.x[2::3])

        self.results.pis = list(pis / pis.sum())
        self.results.sigmas = self.results.x[1::3]
        self.results.mus = self.results.x[0::3]

        return res


class EM(Fitting):
    """Expectation minimization class to estimate parameters of GMM

    .. plot::
        :width: 50%
        :include-source:

        from biokit.stats.mixture import GaussianMixture, EM
        m = GaussianMixture(mu=[-1,1], sigma=[0.5,0.5], mixture=[0.2,0.8])
        em = EM(m.data)
        em.estimate(k=2)
        em.plot()

    """
    def __init__(self, data, model=None, max_iter=100):
        """.. rubric:: constructor

        :param data:
        :param model: not used. Model is the :class:`GaussianMixtureModel` but
            could be other model.
        :param int max_iter: max iteration for the minization

        """
        super(EM, self).__init__(data, k=2) # default is k=2
        self.max_iter = max_iter

    #@do_profile()
    def estimate(self, guess=None, k=2):
        """

        :param list guess: a list to provide the initial guess. Order is mu1, sigma1,
            pi1, mu2, ...
        :param int k: number of models to be used.
        """
        #print("EM estimation")
        self.k = k
        # Initial guess of parameters and initializations
        if guess is None:
            # estimate the mu/sigma/pis from the data
            guess = self.get_guess()

        mu = np.array(guess[0::3])
        sig = np.array(guess[1::3])
        pi_ = np.array(guess[2::3])
        N_ = len(pi_)

        gamma = np.zeros((N_, int(self.size)))
        N_ = np.zeros(N_)
        p_new = guess

        # EM loop
        counter = 0
        converged = False

        self.mus = []

        while not converged:
        # Compute the responsibility func. and new parameters
            for k in range(0, self.k):
                # unstable if eslf.model.pdf is made of zeros

                #self.model.pdf(self.data, p_new,normalise=False).sum()!=0:
                gamma[k, :] = pi_[k] * ss.norm.pdf(self.data, mu[k], sig[k])
                gamma[k, :] /= (self.model.pdf(self.data, p_new, normalise=False))
                """else:
                    gamma[k, :] = pi_[k]*pylab.normpdf(self.data, mu[k],
                        sig[k])/(self.model.pdf(self.data, p_new,
                            normalise=False)+1e-6)
                """
                N_[k] = gamma[k].sum()
                mu[k] = np.sum(gamma[k]*self.data)/N_[k]
                sig[k] = pylab.sqrt( np.sum(gamma[k]*(self.data - mu[k])**2)/N_[k] )
                pi_[k] = N_[k] /  self.size

            self.results = {'x': p_new, 'nfev':counter, 'success': converged}

            p_new = []
            for this in range(self.k):
                p_new.extend([mu[this], sig[this], pi_[this]])

            #p_new = [(mu[x], sig[x], pi_[x]) for x in range(0, self.k)]
            #p_new = list(pylab.flatten(p_new))

            self.status = True
            try:
                assert abs(N_.sum() - self.size)/self.size < 1e-6
                assert abs(pi_.sum() - 1) < 1e-6
            except:
                print("issue arised at iteration %s" % counter)
                self.debug = {'N':N_, 'pis':pi_}
                self.status = False
                break

            self.mus.append(mu)

            # Convergence check
            counter += 1
            converged = counter >= self.max_iter

        self.gamma = gamma

        if self.status is True:
            self.results = {'x': p_new, 'nfev':counter, 'success': converged}


        self.results = AttrDict(**self.results)
        self.results.mus = self.results.x[0::3]
        self.results.sigmas = self.results.x[1::3]
        self.results.pis = self.results.x[2::3]

        log_likelihood = self.model.log_likelihood(self.results.x, self.data)
        self.results.AIC = criteria.AIC(log_likelihood, k, logL=True)

        self.results.log_likelihood = log_likelihood
        self.results.AIC = criteria.AIC(log_likelihood, self.k, logL=True)
        self.results.AICc = criteria.AICc(log_likelihood, self.k, self.data.size, logL=True)
        self.results.BIC = criteria.BIC(log_likelihood, self.k, self.data.size, logL=True)

    def plot(self, model_parameters=None, **kwargs):
        """ Take a list of dictionnaries with models parameters to plot
        predicted models. If user doesn't provide parameters, the standard
        plot function from fitting is used.

        Example:
            model_parameters=[{"mu": 5, "sigma": 0.5, "pi": 1}]
        """
        if not model_parameters:
            return super(EM, self).plot(**kwargs)
        # Set parameters with the dictionnary
        self.k = len(model_parameters)
        self.results = AttrDict()
        self.results.mus = [model["mu"] for model in model_parameters]
        self.results.sigmas = [model["sigma"] for model in model_parameters]
        self.results.pis = [model["pi"] for model in model_parameters]
        parms_keys = ("mu", "sigma", "pi")
        self.results.x = [model[key] for model in model_parameters for key in
                          parms_keys]
        return super(EM, self).plot(**kwargs)


class AdaptativeMixtureFitting(object):
    """Automatic Adaptative Mixture Fitting

    .. plot::
        :width: 80%
        :include-source:

        from biokit.stats.mixture import AdaptativeMixtureFitting, GaussianMixture
        m = GaussianMixture(mu=[-1,1], sigma=[0.5,0.5], mixture=[0.2,0.8])
        amf = AdaptativeMixtureFitting(m.data)
        amf.run(kmin=1, kmax=6)
        amf.diagnostic(k=amf.best_k)


    """
    # TODO: propose to use EM or Minimization.
    def __init__(self, data, method='Nelder-Mead'):
        self.fitting = GaussianMixtureFitting(data, method=method)
        self.verbose = True

    def run(self, kmin=1, kmax=6, criteria='AICc'):
        self.all_results = {}
        self.x = []

        for k in range(kmin, kmax+1):
            self.fitting.estimate(k=k)

            # here criteria does not matter. if one fails, all fail
            if self.fitting.results['AIC'] != 1000:
                self.x.append(k)
                self.all_results[k] = self.fitting.results.copy()

        m = np.array([self.all_results[x][criteria] for x in self.x]).min()
        index = np.array([self.all_results[x][criteria] for x in self.x]).argmin()
        if self.verbose:
            print('Found min ', m, 'for k  ', self.x[index])
        self.best_k = self.x[index]
        self.min_value = m


    def plot(self, criteria='AICc'):

        pylab.plot(self.x, [self.all_results[x]['BIC'] for x in self.x], 'o-', label='BIC')
        pylab.plot(self.x, [self.all_results[x]['AIC'] for x in self.x], 'o-',  label='AIC')
        pylab.plot(self.x, [self.all_results[x]['AICc'] for x in self.x], 'o-', label='AICc')

        m = np.array([self.all_results[x][criteria] for x in self.x]).min()
        # [exp((m-this[criteria])/2) for this in amf.all_results]

        pylab.axhline(m - pylab.log(0.9)*2, color='k', label='90% equi-probability', alpha=0.9)
        pylab.axhline(m - pylab.log(0.5)*2, color='k', label='50% equi-probability', alpha=0.5)
        pylab.axhline(m - pylab.log(0.3)*2, color='k', label='30% equi-probability', alpha=0.3)

        pylab.legend()


    def diagnostic(self, kmin=1, kmax=8, k=None, ymax=None):
        self.run(kmin=kmin, kmax=kmax);
        pylab.clf()
        pylab.subplot(3,1,2);
        self.plot()
        mf = GaussianMixtureFitting(self.fitting.data)
        if k is None:
            mf.estimate(k=self.best_k)
        else:
            mf.estimate(k=k)
        pylab.subplot(3,1,1)
        mf.plot()
        if ymax is not None:
            pylab.ylim([0, ymax])

        pylab.subplot(3,1,3)
        min_value = np.array([self.all_results[x]['AICc'] for x in self.x]).min()
        pylab.plot(self.x, [pylab.exp((min_value-self.all_results[k]['AICc'])/2)
            for k in  self.x], 'o-', label='AICc')
        min_value = np.array([self.all_results[x]['AIC'] for x in self.x]).min()
        pylab.plot(self.x, [pylab.exp((min_value-self.all_results[k]['AIC'])/2)
            for k in  self.x], 'o-', label='AIC')

        pylab.xlabel('probability of information loss (based on AICc')
        pylab.legend()








# -*- python -*-
#
#  This file is part of biokit software
#
#  Copyright (c) 2014 - EBI-EMBL
#
#  File author(s): Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: http://github.com/cellnopt
#
##############################################################################
"""Code use by pyper module original version available from pyper pacakge on pip"""

__all__ = ["r4python"]

r4python = r"""

.getRvalue4Python__ <- function(x, use_dict=NULL) {

    headstr <- 'numpy.array('
    tailstr <- ')'

    SpecialLocs <- function(x) { # find locations of special values: NULL, NA, NaN, Inf
        rlt <- list()

            idx <- which(is.null(x) | is.na(x))
            if (length(idx) > 0) rlt$None <- idx
            idx <- which(is.nan(x))
            if (length(idx) > 0) rlt$numpy.NaN <- idx
            idx <- which(is.infinite(x))
            if (length(idx) > 0) {
                v <- x[idx]
                iidx <- which(v > 0)
                if (length(iidx) > 0) rlt$numpy.Inf <- idx[iidx]
                iidx <- which(v < 0)
                if (length(iidx) > 0) rlt['-numpy.Inf'] <- idx[iidx]
                }

        return(rlt)
        }


    SpecialVals <- function(x, valoc) {
        for (val in names(valoc)) x[valoc[[val]]] <- val
        return(x)
        }

    NullStr <- function(x) 'None'

    VectorStr <- function(x) {
        #nms <- names(x)
        #if (!is.null(nms) &&  length(nms)>0) return(ListStr(as.list(x)))
        complx <- is.complex(x)
        special_locs <- SpecialLocs(x)
        if (is.character(x)) {
            x <- gsub('\\\\', '\\\\\\\\', x)
            x <- gsub('"', '\\\\"', x)
            x <- paste('"', x, '"', sep='') }
        else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
        if (length(special_locs) > 0) x <- SpecialVals(x, special_locs)
        if (length(x)==1) x <- paste(x) # convert to character using paste, "gettext", or "as.character"
        else x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
        if (complx) x <- gsub('i', 'j', x)
        return(x)
    }

    MatrixStr <- function(x) {
        complx <- is.complex(x)
        special_locs <- SpecialLocs(x)
        if (is.character(x)) x <- matrix(paste('"', x, '"', sep=''), nrow=nrow(x))
        else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
        if (length(special_locs) > 0) x <- SpecialVals(x, special_locs)
        x <- apply(x, 1, function(r) paste('[', paste(r, collapse=','), ']', sep=''))
        x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
        if (complx) x <- gsub('i', 'j', x)
        return(x)
    }

    ArrayStr <- function(x) {
        complx <- is.complex(x)
        ndim <- length(dim(x))
        if (ndim == 1) return(VectorStr(x))
        if (ndim == 2) return(MatrixStr(x))
        # ndim >= 3
        if (is.character(x)) x <- array(paste('"', x, '"', sep=''), dim=dim(x))
        else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
        # do col first
        x <- apply(x, seq(dim(x))[-2], function(r) paste('[', paste(r, collapse=','), ']', sep=''))
        for (i in seq(ndim-2))
            x <- apply(x, seq(dim(x))[-1], function(r) paste('[', paste(r, collapse=','), ']', sep=''))
        x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
        if (complx) x <- gsub('i', 'j', x)
        return(x)
    }

    DataFrameStr <- function(x) {
        if (ncol(x) == 0) {
            return('pandas.DataFrame()')
            }

            cnms <- colnames(x) # get column names
            ctp <- list()
            for (i in seq(x)) {
                xi <- as.vector(x[[i]])
                special_locs <- SpecialLocs(xi)
                if (is.character(xi)) {
                    ctp[i] <- sprintf('("%s", "|S%d")', cnms[i], if (length(xi) > 0) max(nchar(xi)) else 0 )
                    # Hack TC July 2017 S%d above was sometime replaced by SNA
                    # in Python3. Let us use object for now
                    ctp[i] <- sprintf('("%s", "object")', cnms[i], 50 )

                    xi <- paste('"', xi, '"', sep='') }
                else if (is.logical(xi)) {
                    xi <- ifelse(xi, 'True', 'False')
                    ctp[i] <- paste('("', cnms[i], '", "<?")' ) }
                else if (is.integer(xi)) {
                    xi <- paste(xi)
                    ctp[i] <- paste('("', cnms[i], '", "<q")' ) }
                else if (is.double(xi)) {
                    xi <- paste(xi)
                    ctp[i] <- paste('("', cnms[i], '", "<g")' ) }
                else if (is.complex(xi)) {
                    xi <- gsub('i', 'j', paste(xi))
                    ctp[i] <- paste('("', cnms[i], '", "<G")') }
                if (length(special_locs) > 0) xi <- SpecialVals(xi, special_locs)
                if (nrow(x) > 0) x[[i]] <- xi }
            tailstr <- paste(', dtype=[', paste(ctp, collapse=','), ']', tailstr, sep='')

        x <- as.matrix(x)
        x <- apply(x, 1, function(r) paste('(', paste(r, collapse=','), if(length(r)<2) ',)' else ')', sep=''))
        x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
        x <- paste('pandas.DataFrame(', x, ')', sep='')
        return(x)
    }

    ListStr <- function(x) {
        nms <- names(x) # get column names
        x <- sapply(x, Str4Py)
        return(zipVecWithName(x, nms))
    }

    zipVecWithName <- function(x, nms) {
        if (!is.null(nms) &&  length(nms)>0) {
            nms <- paste('"', nms, '"', sep='')
            x <- sapply(seq(nms), function(i) paste('(', nms[i], ',', x[i], ')') )
            if (identical(use_dict, TRUE)) x <- paste('dict([', paste(x, collapse=','), '])', sep='')
            else if (identical(use_dict, FALSE))  x <- paste('[', paste(x, collapse=','), ']', sep='')
            else { # should be NULL or something else
                if (any(duplicated(nms))) x <- paste('[', paste(x, collapse=','), ']', sep='')
                else x <- paste('dict([', paste(x, collapse=','), '])', sep='') } }
        else x <- paste('[', paste(x, collapse=','), ']', sep='')
        return(x)
    }

    Str4Py <- function(x) {
        # no considering on NA, Inf, ...
        # use is.XXX, typeof, class, mode, storage.mode, sprintf
        if (is.factor(x)) x <- as.vector(x)
        rlt <- {
            if (is.null(x)) NullStr(x)
            else if (is.vector(x) && !is.list(x)) VectorStr(x)
            else if (is.matrix(x) || is.array(x)) ArrayStr(x)
            else if (is.data.frame(x)) DataFrameStr(x)
            else if (is.list(x)) ListStr(x)
            else if (is.object(x)) 'None'
            else Str4Py(as.character(x)) }
            # other objects will be convert to character (instead of NullStr), or use "gettext"
        return(rlt)
    }
    Str4Py(x)
    }

    # initalize library path for TCL/TK based environment on Windows, e.g. Python IDLE
    .addLibs <- function() {
        ruser <- Sys.getenv('R_USER')
        userpath <- Sys.getenv('R_LIBS_USER')
        libpaths <- .libPaths()
        for (apath in userpath) {
            if (length(grep(apath, libpaths)) > 0) next
            if (file.exists(apath)) .libPaths(apath)
            else {
                d <- '/Documents'
                if (substr(ruser, nchar(ruser)-nchar(d)+1, nchar(ruser)) != d) {
                    apath <- paste(ruser,d, substr(apath, nchar(ruser)+1, nchar(apath)), sep='')
                    if (file.exists(apath)) .libPaths(apath)
                }
            }
        }
    }

    if(identical(.Platform$OS.type, 'windows')) .addLibs()
    rm(.addLibs)

"""



"""

Gaussian Mixture model
=========================

Fit a Gaussian Mixture Models (2 distributions) using brute force optimisation method.

"""

from biokit.stats.mixture import GaussianMixture, GaussianMixtureFitting
m = GaussianMixture(mu=[-1,1], sigma=[0.5,0.5], mixture=[0.2,0.8])
mf = GaussianMixtureFitting(m.data)
mf.estimate(k=2)
mf.plot()




import string
import collections
import pylab

__all__ = ['Sequence']


class Sequence(object):
    """Common data structure to all sequences (e.g., :meth:`~biokit.sequence.dna.DNA`)

    A sequence is a string contained in the :attr:`_data`. If you manipulate this 
    attribute, you should also changed the :attr:`_N` (length of the string) and 
    set :attr:`_counter` to  None.

    Sequences can be concatenated easily. You can also add a string or numpy 
    array or pandas time series to an existing sequence::

        d1 = Sequence('ACGT')
        d2 = Sequence('ACGT')

    Note that there is a :meth:`check` method, which is not called during the 
    instanciation but is called when adding sequences together. Each type of 
    sequence (e.g., Sequence, DNA, RNA) has its own symbols. So you cannot add 
    a DNA sequence with a RNA sequence for instance. Those are valid operation::

        >>> d1 = Sequence('ACGT')
        >>> d1 += 'AAAA'
        >>> d1 + d1
        >>> "AAAA" + d1

    """
    def __init__(self, data=''):
        # initialise before filling _data attribute
        self._checked = False

        if isinstance(data, str):
            self._data = data
        elif isinstance(data, Sequence):
            self._data = data._data
            self._checked = data._checked
        elif self._looks_like_a_sequence(data) is True:
            self._data = data._data
            self._checked = data._checked
        else:
            # assume it is a list or numpy array or pandas TimeSeries
            self._data = "".join(data)

        self._N = len(self._data)
        self._counter = None
        try:
            #python2
            self.symbols = string.punctuation + string.letters
        except:
            # python3
            self.symbols = string.punctuation + string.ascii_letters

        self._type = 'Sequence'

    def _looks_like_a_sequence(self, this):
        # if it looks like a sequence, let us assume it is a sequence
        if hasattr(this, 'symbols') and hasattr(this, '_data') and\
            hasattr(this, '_checked'):
            return True
        else:
            return False

    def _get_N(self):
        return self._N
    N = property(_get_N)

    def __len__(self):
        return self._N

    def _get_sequence(self):
        return self._data[:]
    sequence = property(_get_sequence,
            doc="returns a copy of the sequence")

    def _get_count(self):
        if self._counter is None:
            self._counter = collections.Counter(self._data)
        return self._counter
    counter = property(_get_count, doc="return counter of the letters")

    def histogram(self):
        pylab.clf()
        import pandas as pd
        pd.Series(self.counter).plot(kind='bar')

    def pie(self):
        pylab.clf()
        keys = self.counter.keys()
        labels = dict([(k,float(self.counter[k])/len(self)) for k in keys])

        pylab.pie([self.counter[k] for k in keys], labels=[k +
            ':'+str(labels[k]) for k in keys])

    def hamming_distance(self, other):
        """Return hamming distance between this sequence and another sequence

        The Hamming distance between s and t, denoted dH(s,t), is the number of
        corresponding symbols that differ in s and t.

        ::

            >>> d1 = 'GAGCCTACTAACGGGAT'
            >>> d2 = 'CATCGTAATGACGGCCT'
            >>> s = Sequence(d1)
            >>> s.hamming_distance(d2)
            7
        """
        # TODO:: convert to appropriate sequence.
        return sum(1 for x,y in zip(self._data, other._data) if x!=y)

    def upper(self):
        """convertes sequence string to uppercase (inplace)"""
        self._data = self._data.upper()

    def lower(self):
        """convertes sequence string to lowercase (inplace)"""
        self._data = self._data.lower()

    def _check_sequence(self):
        """checks that characters are valid symbols"""
        for i, x in enumerate(self._data):
            if x not in self.symbols:
                raise ValueError("found invalid symbol %s at position %s" % (x,i))
        self._checked = True

    def __repr__(self):
        if self._N > 10:
            return "%s: %s ... (length %s) " % (self._type, self.sequence[0:10], 
                    self._N)
        else:
            return "%s: %s (length %s) " % (self._type, self.sequence, self._N)

    def __str__(self):
        if self._N > 10:
            return "%s: %s ... (length %s) " % (self._type,self.sequence[0:10], 
                    self._N)
        else:
            return "%s: %s (length %s) " % (self._type, self.sequence, self._N)

    def __convert_to_compat(self, other):
        from biokit.sequence.rna import RNA
        from biokit.sequence.dna import DNA

        if isinstance(self, RNA):
            other = RNA(other)
        elif isinstance(self, DNA):
            other = DNA(other)
        elif isinstance(self, Sequence):
            other = Sequence(other)
        # if self._check is True:
        #     other._check_sequence()
        return other

    def __add__(self, other):
        # input may be a string or list, in which case we need to convert to a sequence
        if isinstance(other, Sequence) is False:
            other = self.__convert_to_compat(other)
        elif type(other) != type(self):
            raise TypeError('incompatible sequences %s versus %s' % (type(other), type(self)))

        # now let us add the 2 sequences
        return self.__convert_to_compat(self._data + other._data)

    def __radd__(self, other):
        """operator other + self"""
        if isinstance(other, Sequence) is False:
            other = self.__convert_to_compat(other)
        elif type(other) != type(self):
            raise TypeError('incompatible sequences %s versus %s' % (type(other), type(self)))

        # now let us add the 2 sequences
        return self.__convert_to_compat(other._data + self._data)

    def __iadd__(self, other):
        if isinstance(other, Sequence) is False:
            other = self.__convert_to_compat(other)
        elif type(other) != type(self):
            raise TypeError('incompatible sequences %s versus %s' % (type(other), type(self)))

        # now let us add the 2 sequences
        self._data += other._data
        self._N = self._N + other._N
        return self

    def __eq__(self, other):
        if isinstance(other, str):
            return self._data == other
        else:
            #assume this is a sequence:
            return self._data == other._data



"""2D histograms



"""
import pylab
import pandas as pd
import numpy as np
from .core import VizInput2D


__all__ = ["Hist2D"]


class Hist2D(VizInput2D):
    """2D histogram


    .. plot::
        :include-source:
        :width: 80%

        from numpy import random
        from biokit.viz import hist2d
        X = random.randn(10000)
        Y = random.randn(10000)
        h = hist2d.Hist2D(X,Y)
        h.plot(bins=100, contour=True)

    """
    def __init__(self, x, y=None, verbose=False):
        """.. rubric:: constructor

        :param x: an array for X values. See :class:`~biokit.viz.core.VizInput2D` for details.
        :param y: an array for Y values. See :class:`~biokit.viz.core.VizInput2D` for details.

        """
        super(Hist2D, self).__init__(x=x, y=y, verbose=verbose)

    def plot(self, bins=100, cmap="hot_r", fontsize=10, Nlevels=4,
        xlabel=None, ylabel=None, norm=None, range=None, normed=False,
        colorbar=True, contour=True, grid=True, **kargs):
        """plots histogram of mean across replicates versus coefficient variation

        :param int bins: binning for the 2D histogram (either a float or list
            of 2 binning values).
        :param cmap: a valid colormap (defaults to hot_r)
        :param fontsize: fontsize for the labels
        :param int Nlevels: must be more than 2
        :param str xlabel: set the xlabel (overwrites content of the dataframe)
        :param str ylabel: set the ylabel (overwrites content of the dataframe)
        :param norm: set to 'log' to show the log10 of the values.
        :param normed: normalise the data
        :param range: as in pylab.Hist2D : a 2x2 shape [[-3,3],[-4,4]]
        :param contour: show some contours (default to True)
        :param bool grid: Show unerlying grid (defaults to True)

        If the input is a dataframe, the xlabel and ylabel will be populated
        with the column names of the dataframe.

        """
        X = self.df[self.df.columns[0]].values
        Y = self.df[self.df.columns[1]].values
        if len(X) > 10000:
            print("Computing 2D histogram. Please wait")

        pylab.clf()
        if norm == 'log':
            from matplotlib import colors
            res = pylab.hist2d(X, Y, bins=bins, normed=normed,
               cmap=cmap, norm=colors.LogNorm())
        else:
            res = pylab.hist2d(X, Y, bins=bins, cmap=cmap,
                    normed=normed, range=range)

        if colorbar is True:
            pylab.colorbar()

        if contour:
            try:
                bins1 = bins[0]
                bins2 = bins[1]
            except:
                bins1 = bins
                bins2 = bins

            X, Y = pylab.meshgrid(res[1][0:bins1], res[2][0:bins2])
            if contour:
                if res[0].max().max() < 10 and norm == 'log':
                    pylab.contour(X, Y, res[0].transpose())
                else:
                    levels = [round(x) for x in
                            pylab.logspace(0, pylab.log10(res[0].max().max()), Nlevels)]
                    pylab.contour(X, Y, res[0].transpose(), levels[2:])
                #pylab.clabel(C, fontsize=fontsize, inline=1)

        if ylabel is None:
            ylabel = self.df.columns[1]
        if xlabel is None:
            xlabel = self.df.columns[0]

        pylab.xlabel(xlabel, fontsize=fontsize)
        pylab.ylabel(ylabel, fontsize=fontsize)

        if grid is True:
            pylab.grid(True)

        return res





"""Heatmap and dendograms"""
import warnings
import scipy.cluster.hierarchy as hierarchy
import scipy.spatial.distance as distance
import easydev


__all__ = ['Linkage']


class Linkage(object):
    """Linkage used in other tools such as Heatmap"""

    def __init__(self):
        """.. rubric:: constructor

        :param data: a dataframe or possibly a numpy matrix.

        """
        pass

    def check_metric(self, value):
        return
        from biokit.viz.commons import valid_metrics
        easydev.check_param_in_list(value, valid_metrics)

    def check_method(self, value):
        # None is possible
        # in R, in addition to single, complete, average, centroid, 
        # median and ward
        # there are  ward.D, wardD2 and mcquitty
        # default is complete
        return
        from biokit.viz.commons import valid_methods
        easydev.check_param_in_list(str(value), valid_methods)

    def linkage(self, df, method, metric):
        self.check_metric(metric)
        self.check_method(method)
        d = distance.pdist(df)
        D = distance.squareform(d)

        # hierarchy.ClusterWarning
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            Y = hierarchy.linkage(D, method=method, metric=metric)
            return Y



# -*- python -*-
#
#  This file is part of biokit software
#
#  Copyright (c) 2014-
#
#  File author(s): Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/biokit
#
##############################################################################
import sys
from types import *

from easydev import check_param_in_list

import pandas
import numpy


__all__ = ["BoolStr", "ReprStr", "FloatStr", "LongStr", "ComplexStr", "UniStr",
"ByteStr", "SeqStr", "getVec", "NumpyNdarrayStr", "PandasDataFrameStr",
"PandasSerieStr", "OtherStr", "Str4R" ]


if sys.version < '3.0':
    _mystr = _mybytes = lambda s: s
    _in_py3 = False
else:
    from functools import reduce
    long, basestring, unicode = int, str, str
    _mybytes = lambda s: bytes(s, 'utf8')  # 'ascii')
    _mystr = lambda s: str(s, 'utf8')
    _in_py3 = True

def BoolStr(obj):
    return(obj and 'TRUE' or 'FALSE')


def ReprStr(obj):
    return(repr(obj))


def FloatStr(f):
    if f is numpy.NaN or f is numpy.nan:
        return('NaN') # or 'NA'
    if pandas.isnull(f):
        return('NaN')
    if numpy.isposinf(f):
        return('Inf')
    if numpy.isneginf(f):
        return('-Inf')
    return(repr(f))


def LongStr(obj):
    rv = repr(obj)
    if rv[-1] == 'L':
        rv = rv[:-1]
    return(rv)


def ComplexStr(obj):
    return(repr(obj).replace('j', 'i'))


def UniStr(obj):
    return(repr(obj.encode('utf8')))


def ByteStr(obj):
    return(repr(obj)[1:])
    #return obj.decode()


def SeqStr(obj, head='c(', tail=')', enclose=True):
    if not enclose: # don't add head and tail
        return(','.join(map(Str4R, obj)))
    if not obj:
        return(head + tail)
    # detect types
    if isinstance(obj, set):
        obj = list(obj)
    obj0 = obj[0]
    tp0 = type(obj0)
    simple_types = [str, bool, int, long, float, complex]
    num_types = [int, long, float, complex]
    is_int = tp0 in (int, long) # token for explicit converstion to integer in R since R treat an integer from stdin as double
    if tp0 not in simple_types:
        head = 'list('
    else:
        tps = isinstance(obj0, basestring) and [StringType] or isinstance(obj0, bool) and [BooleanType] or num_types
        for i in obj[1:]:
            tp = type(i)
            if tp not in tps:
                head = 'list('
                is_int = False
                break
            elif is_int and tp not in (int, long):
                is_int = False
    # convert
    return((is_int and 'as.integer(' or '') + head + ','.join(map(Str4R, obj)) + tail + (is_int and ')' or ''))


def DictStr(obj):
    return('list(' + ','.join(['%s=%s' % (Str4R(a[0]), Str4R(a[1])) for a in obj.items()]) + ')')

# 'b':boo
# lean, 'i':integer, 'u':unsigned int, 'f':float, c complex-float
# 'S'/'a':string, 'U':unicode, 'V':raw data. 'O':string?
_tpdic = {'i':'as.integer(c(%s))', 'u':'as.integer(c(%s))', 'f':'as.double(c(%s))', 'c':'as.complex(c(%s))',
        'b':'c(%s)', 'S':'c(%s)', 'a':'c(%s)', 'U':'c(%s)', 'V':'list(%s)', 'O':'as.character(c(%s))'}
def getVec(ary):
    # used for objects from numpy and pandas
    tp = ary.dtype.kind
    if len(ary.shape) > 1:
        ary = ary.reshape(reduce(lambda a,b=1: a*b, ary.shape))
    ary = ary.tolist()
    if tp != 'V':
        return(_tpdic.get(tp, 'c(%s)') % SeqStr(ary, enclose=False))
    # record array
    ary = list(map(SeqStr, ary)) # each record will be mapped to vector or list
    # use str here instead of repr since it has already been converted to str by SeqStr
    return(_tpdic.get(tp, 'list(%s)') % (', '.join(ary))) 


def NumpyNdarrayStr(obj):
    shp = obj.shape
    if len(shp) == 1: # to vector
        tp = obj.dtype
        if tp.kind != 'V':
            return(getVec(obj))
        # One-dimension record array will be converted to data.frame
        def mapField(f):
            ary = obj[f]
            tp = ary.dtype.kind
            return('"%s"=%s' % (f, _tpdic.get(tp, 'list(%s)') % SeqStr(ary.tolist(), enclose=False)))
        return('data.frame(%s)' % (', '.join(map(mapField, tp.names))))
    elif len(shp) == 2: # two-dimenstion array will be converted to matrix
        return('matrix(%s, nrow=%d, byrow=TRUE)' % (getVec(obj), shp[0]))
    else: # to array
        dim = list(shp[-2:]) # row, col
        dim.extend(shp[-3::-1])
        newaxis = list(range(len(shp)))
        newaxis[-2:] = [len(shp)-1, len(shp)-2]
        return('array(%s, dim=c(%s))' % (getVec(obj.transpose(newaxis)), repr(dim)[1:-1]))


def PandasSerieStr(obj):
    return('data.frame(%s=%s, row.names=%s)' % (obj.name, getVec(obj.values), getVec(obj.index)))


def PandasDataFrameStr(obj):
    # DataFrame will be converted to data.frame, have to explicitly name columns
    #return 'data.frame(%s, row.names=%s)' % (', '.join(map(lambda a,b=obj:a+'='+getVec(obj[a]), obj)), getVec(obj.index))
    s = ', '.join(map(lambda a,b=obj: '"%s"=%s' % (str(a), getVec(obj[a])), obj))
    return('data.frame(%srow.names=%s)' % (s and s+', ', getVec(obj.index)))
    s = ''
    for col in obj:
        s = s + col + '=' + getVec(obj[col]) + ', '
    # print 'data.frame(%s row.names=%s)' % (s, getVec(obj.index))
    return('data.frame(%s row.names=%s)' % (s, getVec(obj.index)))


def OtherStr(obj):
    if hasattr(obj, '__iter__'):  # for iterators
        if hasattr(obj, '__len__') and len(obj) <= 10000:
            return(SeqStr(list(obj)))
        else:  # waiting for better solution for huge-size containers
            return(SeqStr(list(obj)))
    return(repr(obj))


str_func = {
        type(None): 'NULL',
        bool: BoolStr,
        long: LongStr,
        int: repr,
        float: FloatStr,
        complex: ComplexStr,
        unicode: UniStr,
        str: repr,
        list: SeqStr,
        tuple: SeqStr,
        set: SeqStr,
        frozenset: SeqStr,
        dict: DictStr} # str will override uncode in Python 3

base_tps = [type(None), bool, int, long, float, complex, str, unicode, list,
        tuple, set, frozenset, dict] # use type(None) instead of NoneType since
        #the latter cannot be found in the types module in Python 3

str_func[numpy.ndarray] = NumpyNdarrayStr
base_tps.append(numpy.ndarray)

str_func.update({pandas.Series: PandasSerieStr, pandas.DataFrame: PandasDataFrameStr})
base_tps.extend([pandas.Series, pandas.DataFrame])
base_tps.reverse()

if _in_py3:
    base_tps.append(bytes)
    str_func[bytes] = ByteStr


def Str4R(obj):
    """
    convert a Python basic object into an R object in the form of string.
    """
    # for objects known by PypeR
    if type(obj) in str_func:
        return(str_func[type(obj)](obj))
    # for objects derived from basic data types
    for tp in base_tps:
        if isinstance(obj, tp):
            return(str_func[tp](obj))
    # for any other objects
    return(OtherStr(obj))






def download_fasta(identifier, tag=None, service=None):
    """Retrieve FASTA from ENA, or UniProt, or EUtils

    :param identifier: valid identifier to retrieve FASTA (from ENA)
    :param tag. filename without extension (.fa)
    :param service: one of ENA, UniProt, EUtils to force the usage of one
        service (instead of 3). This is fasta since it avoids trying a service
        that may fail. See details below

    If the identifier is not found in ENA, we assume it is a protein sequence
    from uniprot. It it fails, we assume that the identifier is a NCBI
    identifier so we try EUtils service. If you know that the service is EUtils, 
    set service=EUtils which will skip the ENA and UniProt trials (that will
    fail).
    """
    from bioservices import ENA, UniProt, EUtils

    if service is None or service == "ENA":
        ena = ENA()
        data = ena.get_data(identifier, 'fasta')
        if isinstance(data, int) and data == 400 or "not supported" in data.decode():
            data = None

    # if not ENA, let us try UniProt
    if data is None and (service is None or service == "UniProt"):
        u = UniProt()
        data = u.get_fasta(identifier)
        if isinstance(data, int) and data == 400:
            data = None

    if data is None and (service is None or service == "EUtils"):
        ss = EUtils()
        data = ss.EFetch("nuccore", "808039399", rettype="fasta").decode()
        if isinstance(data, int) and data == 400:
            data = None


    if tag is None:
        return data
    else:
        with open("%s.fa" % tag, "w") as fout:
            fout.write(data)


def download_genbank(identifier, tag=None, db="nuccore", rettype="gbwithparts"):
    """

    :param identifier: valid identifier to retrieve from NCBI
    :param tag. filename without extension (.fa)

    ::

        from biokit.tools.downloader import download_genbank
        download_genbank('JB409847')

    """
    from bioservices import EUtils
    eu = EUtils()
    data = eu.EFetch(db=db, id=identifier, rettype=rettype, retmode="text")
    if isinstance(data, int) and data == 400:
        raise ValueError("%s not found on NCBI")

    if tag is None:
        return data.decode()
    else:
        with open("%s.gbk" %  tag, "w") as fout:
            fout.write(data.decode())
